(module setheo
 (include "setheo.sch")
 (export (setheo n)))
 
(define (nconc l1 l2)
  (if (null? l2)
      l1
      (if (null? l1)
	  l2
	  (set-cdr! (last l1) l2))))

(define (last l)
  (if (null? (cdr l))
      l
      (last (cdr l))))

(define (nreverse list)
  (if (null? list)
      list
      (nreverse1 list '() '())))

(define (nreverse1 list prev tmp)
  (set! tmp (cdr list))
  (set-cdr! list prev)
  (if (null? tmp)
      list
      (nreverse1 tmp list '())))

(define (clause-bigger? c1 c2)
  (if (> (clause-length c1) (clause-length c2))
      #t
      (if (< (clause-length c1) (clause-length c2))
	  #f
	  (if (> (clause-complexity c1) (clause-complexity c2))
	      #t
	      #f))))

(define (sort-clauses clauselist)
  (if (or (null? clauselist) (null? (cdr clauselist)))
      clauselist
      (begin
	(select-smallest-clause clauselist (cdr clauselist) '())
	(sort-clauses (cdr clauselist)))))

(define (select-smallest-clause clauselist restlist tmp)
  (if (clause-bigger? (car clauselist) (car restlist))
      (begin
	(set! tmp (car clauselist))
	(set-car! clauselist (car restlist))
	(set-car! restlist tmp)))
  (if (not (null? (cdr restlist)))
      (select-smallest-clause clauselist (cdr restlist) '())))

(define (set-union list1 list2)
 (if (pair? list1)
     (if (element (car list1) list2)
	 (set-union (cdr list1) list2)
	 (begin (set! list2 (cons (car list1) list2))
		(set-union (cdr list1) (list2))))
     list2))

(define (element e list)
 (if (pair? list)
     (if (eq? e (car list)) 
	 list
	 (element e (cdr list)))
     #f));; -------------------------------------------------------
;; PARSER FOR TPTP INPUT FILES
;; -------------------------------------------------------

(define current-clause '())
(define current-clause-nr 0)
(define current-clause-variables '())
(define current-clause-var-occurrences '())
(define current-variable '())
(define current-literal '())
(define current-predicate '())
(define current-pred-nr 1)
(define current-functor '())
(define current-signature '())
(define current-functor-nr 1)
(define predicate-list '())
(define functor-list '())
(define propositional-list '())
(define constant-list '())
(define current-positive-subgoals 0)
(define current-negative-subgoals 0)
(define nr-clauses 0)               
(define nr-non-horn 0)              
(define nr-units 0)                 
(define nr-literals 0)              
(define nr-eq-literals 0)           
(define max-clause-length 0)        
(define nr-predicates 0)            
(define nr-propositionals 0)        
(define max-pred-arity 0)           
(define min-pred-arity 0)           
(define nr-functors 0)              
(define nr-constants 0)             
(define max-func-arity 0)           
(define min-func-arity 0)           
(define nr-variables 0)             
(define nr-singleton-variables 0)   
(define nr-axioms 0)   
(define nr-lemmas 0)   
(define nr-hypotheses 0)   
(define nr-conjectures 0)   

(define tptp-scanner
  (regular-grammar
   ((lowercase (in (#\a #\z)))
    (uppercase (in (#\A #\Z)))
    (numeric (in (#\0 #\9)))
    (special (in #\+ #\- #\_))
    (alphanum (or lowercase uppercase numeric special)))
   ("input_clause" (display-for-debug) 'input_clause)
;;   ("axiom" (display-for-debug) 
;;    (set! nr-axioms (+ nr-axioms 1)) (cons 'type (the-symbol))) 
;;   ("hypothesis" (display-for-debug) 
;;    (set! nr-hypotheses (+ nr-hypotheses 1)) (cons 'type (the-symbol))) 
;;   ("conjecture" (display-for-debug) 
;;    (set! nr-conjectures (+ nr-conjectures 1)) (cons 'type (the-symbol))) 
   ("equal" (display-for-debug)
	      (cons 'symbol (the-symbol)))
   ((: uppercase (* alphanum)) ;; (print (the-string))
			     (cons 'variable (string->symbol (the-string))))
   ((: (or lowercase numeric) (* alphanum)) ;; (print (the-string))
                             (cons 'symbol (string->symbol (the-string))))
   ("++" (display-for-debug) 'pos_sign)
   ("--" (display-for-debug) 'neg_sign)
   (#\( (display-for-debug) 'openpar)
   (#\) (display-for-debug) 'closepar)
   (#\[ (display-for-debug) 'openbracket)
   (#\] (display-for-debug) 'closebracket)
   (#\, (display-for-debug) 'colon)
   (#\. (display-for-debug) 'dot)

;   ((#\%) (context 'comment) (display-for-debug) (ignore))
;   ((context ('comment) (eol (* all))) (display-for-debug) (ignore) (context))

   ((eol (: #\% (* all))) (display-for-debug) (ignore))

   ((+ (in #\space #\tab #\newline)) (display-for-debug) (ignore))))

(define (variable-exists? variable variables)
  (if (null? variables)
      #f
      (if (eq? variable (var-name (car variables)))
	  #t
	  (variable-exists? variable (cdr variables)))))

(define (get-variable variable variables)
  (if (null? variables)
      #f
      (if (eq? variable (var-name (car variables)))
	  (car variables)
	  (get-variable variable (cdr variables)))))

(define (increment-occurences variable variables occurrences)
  (if (null? variables)
      #f
      (if (eq? variable (var-name (car variables)))
	  (set-car! occurrences (+ (car occurrences) 1))
	  (increment-occurences variable (cdr variables) (cdr occurrences)))))

(define (set-copy-of-litlist litlist copy)
  (if (null? litlist)
      #t
      (begin
	(set-subgoal-copy (car litlist) copy)
	(set-copy-of-litlist (cdr litlist) copy))))

(define (compute-clause-complexity clause)
  (let ((cx 0)
	(len (clause-length clause)))
    (do ((i 0 (+ i 1)))
	((= i len) (- cx (clause-nr-variables clause)))
      (set! cx (+ cx (subgoal-complexity (vector-ref (clause-generic clause) i)))))))

(define (term-complexity term)
 (if (or (variable term) (constant term)) 
     1
     (+ 1 (termlist-complexity (cdr term)))))
   
(define (termlist-complexity terms)
  (if (null? terms)
      0
      (+ (term-complexity (car terms)) (termlist-complexity (cdr terms)))))

(define (subgoal-complexity subgoal)
  (let ((terms (subgoal-terms subgoal)))
    (termlist-complexity terms)))

(define tptp-parser
  (lalr-grammar
   (variable symbol input_clause pos_sign neg_sign openpar closepar 
	     openbracket closebracket colon dot) 
   (formula ((clauselist) (set! nr-clauses current-clause-nr)
			  (set! nr-predicates (length predicate-list))
			  (set! nr-propositionals (length propositional-list))
			  (set! nr-functors (length functor-list))
			  (set! nr-constants (length constant-list))
			  (set! max-pred-arity (get-max-pred-arity predicate-list))
			  (set! min-pred-arity (get-min-pred-arity predicate-list))
			  (set! max-func-arity (get-max-func-arity functor-list))
			  (set! min-func-arity (get-min-func-arity functor-list))
			  (set-formula-clauses formula (reverse clauselist))))
   (clauselist
    (() '())
    ((clauselist clause) (cons clause clauselist)))
   (clause
    ((input_clause openpar clausename colon type colon literals closepar dot)
     (set! current-clause (make-vector 12))
     (set-clause-name current-clause clausename)
     (set-clause-type current-clause type)
     (set! current-clause-nr (+ current-clause-nr 1))
     (set-clause-number current-clause current-clause-nr)
     (set-clause-generic current-clause 
			 (list->vector (append literals (list #f))))
     (set-clause-length current-clause (length literals))
     (set! nr-literals (+ nr-literals (clause-length current-clause)))
     (if (> (clause-length current-clause) max-clause-length)
	 (set! max-clause-length (clause-length current-clause)))
     (if (= (clause-length current-clause) 1)
	 (set! nr-units (+ nr-units 1)))
     (set-copy-of-litlist literals current-clause)
     (set-clause-copies current-clause 
			(list (make-clause-copy (clause-generic current-clause) 
						(clause-length current-clause) 
						'() current-clause '())))
     (set-clause-all-copies current-clause (clause-copies current-clause))
     (set-clause-variables current-clause current-clause-variables)
     (set! nr-variables (+ nr-variables (length current-clause-variables)))
     (set! nr-singleton-variables 
	   (+ nr-singleton-variables 
	      (get-nr-current-singletons current-clause-var-occurrences)))
     (set-clause-nr-variables current-clause (length current-clause-variables))
     (set-clause-complexity current-clause 
			    (compute-clause-complexity current-clause))
     (set! current-clause-variables '())
     (set! current-clause-var-occurrences '())
     (if (> current-positive-subgoals 1)
	 (set! nr-non-horn (+ nr-non-horn 1)))
     (set! current-positive-subgoals 0) (set! current-negative-subgoals 0)
     (set-clause-signature current-clause current-signature)
     (set! current-signature '())
     current-clause))
   (literals
    ((openbracket litlist closebracket) litlist))
   (litlist
    ((litlist colon literal) (append litlist (list literal)))
    ((literal) (list literal)))
   (literal
    ((pos_sign atom)
     (set! current-literal (make-vector 8))
     (set-subgoal-sign current-literal '+)
     (set-subgoal-pred current-literal (car atom))
     (set-subgoal-terms current-literal (cdr atom))
     (set-subgoal-links current-literal '())
     (set-subgoal-selected current-literal #f)
     (set-subgoal-next current-literal #f)
     (set! current-positive-subgoals (+ current-positive-subgoals 1))
     (set-subgoal-number current-literal 
              (+ current-negative-subgoals current-positive-subgoals))
     current-literal)
    ((neg_sign atom)
     (set! current-literal (make-vector 8))
     (set-subgoal-sign current-literal '-)
     (set-subgoal-pred current-literal (car atom))
     (set-subgoal-terms current-literal (cdr atom))
     (set-subgoal-links current-literal '())
     (set-subgoal-selected current-literal #f)
     (set-subgoal-next current-literal #f)
     (set! current-negative-subgoals (+ current-negative-subgoals 1))
     (set-subgoal-number current-literal 
              (+ current-negative-subgoals current-positive-subgoals))
     current-literal))
   (atom
    ((symbol) (set! nr-propositionals (+ nr-propositionals 1)) 
	      ;; (print symbol)
	      (if (not (get-predsym symbol 0 predicate-list))
		  (begin
		    (if sgindex
			(begin (set! current-predicate (make-vector 6))
			       (set-predsym-pos-index current-predicate #f)
			       (set-predsym-neg-index current-predicate #f))
			(set! current-predicate (make-vector 4)))
		    (set-predsym-symbol current-predicate symbol)
		    (set-predsym-arity  current-predicate 0)
		    (set-predsym-number current-predicate current-pred-nr)
		    (set-predsym-weight current-predicate 1)
		    (set! current-pred-nr (+ current-pred-nr 1))
		    (set! predicate-list (cons current-predicate predicate-list)))
		  (set! current-predicate (get-predsym symbol 0 predicate-list)))
	      (if (not (get-signature symbol 0 current-signature))
		    (set! current-signature (cons current-predicate 
						  current-signature)))
	      (if (not (member symbol propositional-list))
		  (set! propositional-list (cons symbol propositional-list)))
	      (cons current-predicate '()))
    ((symbol parameters) 
     ;; (print symbol)
     (if (eq? symbol 'equal)
	 (set! nr-eq-literals (+ nr-eq-literals 1)))
     (if (not (get-predsym symbol (length parameters) predicate-list))
	 (begin
	   (if sgindex
	       (begin (set! current-predicate (make-vector 6))
		      (set-predsym-pos-index current-predicate (cons '() 0)) 
		      (set-predsym-neg-index current-predicate (cons '() 0)))
	       (set! current-predicate (make-vector 4)))
	   (set-predsym-symbol current-predicate symbol)
	   (set-predsym-arity  current-predicate (length parameters))
	   (set-predsym-number current-predicate current-pred-nr)
           (set-predsym-weight current-predicate 1)
	   (set! current-pred-nr (+ current-pred-nr 1))
	   (set! predicate-list (cons current-predicate predicate-list)))
	 (set! current-predicate (get-predsym symbol (length parameters) predicate-list)))
     (if (not (get-signature symbol (length parameters) current-signature))
	 (set! current-signature (cons current-predicate current-signature)))
     (cons current-predicate parameters)))
   (term
    ((openpar term closepar) term)
    ((symbol) 
     (if (not (get-functor symbol 0 functor-list))
	 (begin
	   (set! current-functor (make-vector 4))
	   (set-functor-symbol current-functor symbol)
	   (set-functor-arity  current-functor 0)
	   (set-functor-number current-functor current-functor-nr)
	   (set-functor-weight current-functor 1)
	   (set! current-functor-nr (+ current-functor-nr 1))
	   (set! functor-list (cons current-functor functor-list)))
	 (set! current-functor (get-functor symbol 0 functor-list)))
     (if (not (get-signature symbol 0 current-signature))
	 (set! current-signature (cons current-functor current-signature)))
     (if (not (member symbol constant-list))
	 (set! constant-list (cons symbol constant-list)))
     (list current-functor))
    ((variable)
     (if (variable-exists? variable current-clause-variables)
	 (begin (set! current-variable 
		      (get-variable variable current-clause-variables))
		(increment-occurences variable current-clause-variables 
				      current-clause-var-occurrences))
	 (begin
	   (set! current-variable (cons '() (cons #f variable)))
	   (set! current-clause-var-occurrences
		 (cons 1 current-clause-var-occurrences))
	   (set! current-clause-variables 
		 (cons current-variable current-clause-variables))))
      current-variable)
    ((symbol parameters) 
     (if (not (get-functor symbol (length parameters) functor-list))
	 (begin
	   (set! current-functor (make-vector 4))
	   (set-functor-symbol current-functor symbol)
	   (set-functor-arity  current-functor (length parameters))
	   (set-functor-number current-functor current-functor-nr)
	   (set-functor-weight current-functor 1)
	   (set! current-functor-nr (+ current-functor-nr 1))
	   (set! functor-list (cons current-functor functor-list)))
	 (set! current-functor (get-functor symbol (length parameters) functor-list)))
     (if (not (get-signature symbol (length parameters) current-signature))
	 (set! current-signature (cons current-functor current-signature)))
     (append (list current-functor) parameters)))
   (parameters
    ((openpar termlist closepar) termlist))
   (termlist
    ((termlist colon term) (append termlist (list term)))
    ((term) (list term)))
   (type
    ((symbol) 
     (cond ((eq? symbol (string->symbol "axiom"))
            (set! nr-axioms (+ nr-axioms 1)) 'axiom)
	   ((eq? symbol (string->symbol "lemma"))
	    (set! nr-lemmas (+ nr-lemmas 1)) 'lemma)
	   ((eq? symbol (string->symbol "hypothesis"))
	    (set! nr-hypotheses (+ nr-hypotheses 1)) 'hypothesis)
	   ((eq? symbol (string->symbol "conjecture"))
	    (set! nr-conjectures (+ nr-conjectures 1)) 'conjecture))
	   ))
   (clausenamepars
    ((openpar clausenametl closepar) clausenametl))
   (clausenametl
    ((clausenametl colon clausename) (append clausenametl (list clausename)))
    ((clausename) (list clausename)))
   (clausename
    ((openpar clausename closepar) clausename)
    ((symbol) symbol)
    ((variable) variable)
    ((symbol clausenamepars) 
     (append (list symbol) clausenamepars)))))


(define (get-nr-current-singletons occurrences-list)
  (if (null? occurrences-list)
      0
      (if (= (car occurrences-list) 1)
	  (+ (get-nr-current-singletons (cdr occurrences-list)) 1)
	  (get-nr-current-singletons (cdr occurrences-list)))))

(define (get-max-pred-arity predlist)
  (if (null? predlist)
      '-
      (let ((tmp (get-max-pred-arity (cdr predlist))))
	(if (or (eq? tmp '-) (< tmp (predsym-arity (car predlist))))
	    (predsym-arity (car predlist))
	    tmp))))

(define (get-min-pred-arity predlist)
  (if (null? predlist)
      '-
      (let ((tmp (get-min-pred-arity (cdr predlist))))
	(if (or (eq? tmp '-) (> tmp (predsym-arity (car predlist))))
	    (predsym-arity (car predlist))
	    tmp))))

(define (get-max-func-arity funclist)
  (if (null? funclist)
      '-
      (let ((tmp (get-max-func-arity (cdr funclist))))
	(if (or (eq? tmp '-) (< tmp (functor-arity (car funclist))))
	    (functor-arity (car funclist))
	    tmp))))

(define (get-min-func-arity funclist)
  (if (null? funclist)
      '-
      (let ((tmp (get-min-func-arity (cdr funclist))))
	(if (or (eq? tmp '-) (> tmp (functor-arity (car funclist))))
	    (functor-arity (car funclist))
	    tmp))))

(define (get-predsym symbol arity predicate-list)
  (if (null? predicate-list)
      #f
      (if (and (eq? symbol (predsym-symbol (car predicate-list)))
	       (eq? arity (predsym-arity (car predicate-list))))
	  (car predicate-list)
	  (get-predsym symbol arity (cdr predicate-list)))))

(define (get-functor symbol arity functor-list)
  (if (null? functor-list)
      #f
      (if (and (eq? symbol (functor-symbol (car functor-list)))
	       (eq? arity (functor-arity (car functor-list))))
	  (car functor-list)
	  (get-functor symbol arity (cdr functor-list)))))

(define (get-signature symbol arity signature-list)
  (if (null? signature-list)
      #f
      (if (and (eq? symbol (signature-symbol (car signature-list)))
	       (eq? arity (signature-arity (car signature-list))))
	  (car signature-list)
	  (get-signature symbol arity (cdr signature-list)))))

(define (display-statistics)
  (display #"Syntax:\t")
  (display #"Number of clauses:\t")
  (display nr-clauses)
  (display " ( ")
  (display nr-non-horn)
  (display " non-Horn; ")
  (display nr-units)
  (display " unit)")
  (newline)
  (display #"\t")
  (display #"Number of literals:\t")
  (display nr-literals)
  (display " ( ")
  (display nr-eq-literals)
  (display " equality)")
  (newline)
  (display #"\t")
  (display #"Maximal clause size:\t")
  (display max-clause-length)
  (newline)
  (display #"\t")
  (display #"Number of predicates:\t")
  (display nr-predicates)
  (display " ( ")
  (display nr-propositionals)
  (display " propositional; ")
  (display min-pred-arity)
  (display "-")
  (display max-pred-arity)
  (display " arity)")
  (newline)
  (display #"\t")
  (display #"Number of functors:\t")
  (display nr-functors)
  (display " ( ")
  (display nr-constants)
  (display " constant; ")
  (display min-func-arity)
  (display "-")
  (display max-func-arity)
  (display " arity)")
  (newline)
  (display #"\t")
  (display #"Number of variables:\t")
  (display nr-variables)
  (display " ( ")
  (display nr-singleton-variables)
  (display " singleton)")
  (newline))

(define (display-for-debug)
  ;;(display (the-string))
  #t)

(define (reset-parser)
  (set! current-clause '())
  (set! current-clause-nr 0)
  (set! current-clause-variables '())
  (set! current-clause-var-occurrences '())
  (set! current-variable '())
  (set! current-literal '())
  (set! current-predicate '())
  (set! current-pred-nr 1)
  (set! current-functor '())
  (set! current-functor-nr 1) 
  (set! current-signature '())
  (set! predicate-list '())
  (set! functor-list '())
  (set! propositional-list '())
  (set! constant-list '())
  (set! current-positive-subgoals 0)
  (set! nr-clauses 0)
  (set! nr-non-horn 0)
  (set! nr-units 0)
  (set! nr-literals 0)
  (set! nr-eq-literals 0)
  (set! max-clause-length 0)
  (set! nr-predicates 0)
  (set! nr-propositionals 0)
  (set! max-pred-arity 0)
  (set! min-pred-arity 0)
  (set! nr-functors 0)
  (set! nr-constants 0)
  (set! max-func-arity 0)
  (set! min-func-arity 0)
  (set! nr-variables 0)
  (set! nr-singleton-variables 0))
(define tptp-base "/group/TPTP/TPTP-v2.1.0/Problems/all_probs/")
(define tptp-extension ".p")

;(define (read-tptp)
;  (let ((infile (open-input-file "/group/TPTP/TPTP-v2.1.0/Problems/all_probs/ANA002-1.p")))
;    (do ((c #\ ))
;	((eof-object? c))
;      (set! c (read-char infile))
;      (if (not (eof-object? c))
;	  (write-char c)))
;    (close-input-port infile)))

(define (read-tptp)
  (let (infile)
;    (display "Please enter TPTP file name (no extension): ")
;    (set! filename (read-string (current-input-port) 255))
;    (set! filename (substring filename 0 (- (string-length filename) 1)))
    ;;(display filename) (display ": ")
    ;;(display (string-length filename))
;    (set! filename (string-append filename tptp-extension))
    (display filename) (display ": ")
    (if (not (file-exists? filename))
	(begin
	  (display "File not found")
	  (newline)
	  (exit 1)
	  #f)
	(begin (display "Trying to parse")
	       (newline)
	       (set! infile (open-input-file filename))
	       (read/lalrp tptp-parser tptp-scanner infile)
	       (close-input-port infile)
	       (display-statistics)
	       #t
	       ))))
;;;*************************************************************************
;; ------------------------
;; GENERAL GLOBAL VARIABLES
;; ------------------------

(define trail-root (list (cons #f #f)))
(define trail trail-root)
(define v-trail-root (list (vector #f #f #f #f)))
(define v-trail v-trail-root)
(define c-stack-root (list #f))
(define c-stack c-stack-root)
(define actual-c-stack c-stack)
(define eq-trail-root (list #f))
(define eq-trail eq-trail-root)
(define eq-trail-before eq-trail-root)
(define eq-links '())
(define difference-counter 0)
(define filename #f)
(define formula (make-vector 3))
(set-formula-lemmas formula '())
(define start-clauses '())
(define norel #f)
(define global-subgoals #f)
(define actual-copy (make-vector 10))
(define all-global-subgoals #f)
(define current-subgoal #f)
(define current-complexity #f)
(define subgoal-complexity-measure 0)
(define new-complexity #f)
(define current-vars #f)
(define new-vars #f)
(define optimized-unification #f)
(define bound-fail #f)
(define actual-depth 0)
(define free-sig 0)
(define old-free-sig 0)
(define sig-vector #f)
(define sig-trail #f)
(define sig-trail-position 0)
(define old-sig-trail-position 0)
(define proof-output #f)
(define successful-inferences 0)
(define successful-matchings 0)
(define successful-reduction-steps 0)
(define inf 0)
(define nr-reduction-steps 0)
(define nr-matchings 0)
(define failed-inferences 0)
(define constraint-fails 0)
(define total-successful-inferences 0)
(define total-failed-inferences 0)
(define total-constraint-fails 0)
(define index-root #f)
(define index #f)
(define in-index? #t)
(define index-bindings (make-vector 100 #f))
(define max-variables 0) 
(define nth-variable 0)
(define var-assoc-nr? #f)
(define constraints #f)
(define regularity #t)
(define regularity-only-for-subgoals #t)
(define sgindex #f)
(define nr-subgoals 0)
(define index-size 0)

;;;*************************************************************************
;; -----------------------------------------
;; DEFAULTS SETTINGS FOR THE SEARCH STRATEGY
;; -----------------------------------------

(define clause-reordering #f)
(define subgoal-alternation #f)
(define subgoal-reordering #f)
(define global-reordering #f)
(define clause-type-reliable #t)
(define signature-bound #f)
(define equality #f)
(define equal-predecessor-fail #f)
(define depth 1)
(define final-depth 0)
(define weighted-depth #f)
(define weighted-depth-entered 0)
(define wd1 120)
(define wd2 120)
(define wd3 120)
(define matings-pruning #f)

;;;*************************************************************************
;; ---------------------------
;; UPDATE OF DEFAULTS SETTINGS
;; ---------------------------

(define (update-defaults args)
 (if (pair? (cdr args))
     (cond ((equal? (car args) "-d")
            (set! final-depth (string->integer (cadr args)))
	    (update-defaults (cddr args)))
           ((equal? (car args) "-id")
            (set! depth (string->integer (cadr args)))
	    (update-defaults (cddr args)))
           ((equal? (car args) "-noclreord")
            (set! clause-reordering #f)
	    (update-defaults (cdr args)))
           ((equal? (car args) "-wdr")
            (set! weighted-depth #t)
	    (update-defaults (cdr args)))
           ((equal? (car args) "-wd1")
            (set! weighted-depth #t)
            (set! wd1 (string->integer (cadr args)))
            (update-defaults (cddr args)))
           ((equal? (car args) "-wd2")
            (set! weighted-depth #t)
            (set! wd2 (string->integer (cadr args)))
	    (update-defaults (cddr args)))
           ((equal? (car args) "-wd3")
            (set! weighted-depth #t)
            (set! wd3 (string->integer (cadr args)))
	    (update-defaults (cddr args)))
           ((equal? (car args) "-sig")
            (set! signature-bound #t)
            (set! free-sig (string->integer (cadr args)))
	    (update-defaults (cddr args)))
           ((equal? (car args) "-dynsgreord")
            (set! subgoal-reordering (string->integer (cadr args)))
	    (update-defaults (cddr args)))
           ((equal? (car args) "-complexity")
            (set! subgoal-complexity-measure (string->integer (cadr args)))
	    (update-defaults (cddr args)))
           ((equal? (car args) "-singledelay")
            (set! subgoal-alternation 1)
	    (update-defaults (cdr args)))
           ((equal? (car args) "-reg")
            (set! constraints #t)
	    (update-defaults (cdr args)))
           ((equal? (car args) "-multidelay")
            (set! subgoal-alternation 2)
	    (update-defaults (cdr args)))
           ((equal? (car args) "-globreord")
            (set! global-reordering #t)
	    (update-defaults (cdr args)))
           ((equal? (car args) "-mate")
            (set! matings-pruning #t)
	    (update-defaults (cdr args)))
           ((equal? (car args) "-proof")
            (set! proof-output #t)
	    (update-defaults (cdr args)))
           ((equal? (car args) "-sgindex")
            (set! sgindex #t)
	    (update-defaults (cdr args)))
           ((equal? (car args) "-equality")
            (set! equality #t)
	    (update-defaults (cdr args)))
           ((equal? (car args) "-equality")
            (set! equality #t)
	    (update-defaults (cdr args)))
           ((equal? (car args) "-norel")
            (set! norel #t)
	    (update-defaults (cdr args)))
           ((equal? (car args) "-fvector")
            (make-vector (string->integer (cadr args)) #f)
	    (update-defaults (cddr args)))
           ((equal? (car args) "-eqpred")
            (let ((option (string->integer (cadr args))))
             (if (= option 0)
                 (begin (set! equal-predecessor-fail 0)
                        (update-defaults (cdr args)))
                 (begin (set! equal-predecessor-fail 1)
                        (update-defaults (cddr args))))))
	   (else (display "Abort: error in parameter list") 
	         (newline) (exit 1))
)))

;;;*************************************************************************
;; -------------
;; FOR DEBUGGING
;; -------------

(define l1 #f)
(define l2 #f)
(define cl #f)
(define d #f)
(define i 0)


;;;*************************************************************************
;; -------------------------
;; STRUCTURE COPY PROCEDURES
;; -------------------------

(define (copy-termlist termlist)
;; computes a copy of termlist with the variables consistently renamed.
;; Since the trail is used for keeping the variable consistency, 
;; - before each call to copy-termlist, one has to execute (trail-stop)
;; - and afterwards one has to execute (unbind-vars)
;; NOTE: termlist has to be free of ref-cells
 (if (null? termlist)
     '()
     (cons (copy-term (car termlist)) (copy-termlist (cdr termlist)))))

(define (copy-term term)
;; computes a copy of term with the variables consistently renamed.
;; Since the trail is used for keeping the variable consistency, 
;; - before each call to copy-term, one has to execute (trail-stop)
;; - and afterwards one has to execute (unbind-vars)
;; NOTE: term has to be free of ref-cells
 (if (non-variable term)
     (cons (car term) (copy-termlist (cdr term)))
     (if (free-variable term)
         (begin (bind term (cons '() (cons #f (var-name term))))
;(cons (copy-constraints (var-constraints term)))
                (car term))
         (car term))))

; Note: for using static constraints, this function has to be implemented
; and integrated into the copy-term function
(define (copy-constraints constraints)
 (if constraints
     (display "copy-constraints not yet implemented")
     constraints))

(define (make-subgoal-copy subgoal clause-copy)
;; computes a variable-renamed copy of a subgoal.
;; Since the trail is used for keeping the variable consistency, 
;; - before each call to make-subgoal-copy, one has to execute (trail-stop)
;; - and afterwards one has to execute (unbind-vars)
 (let ((sg (make-vector 8)))
  (set-subgoal-sign sg (subgoal-sign subgoal))
  (set-subgoal-pred sg (subgoal-pred subgoal))
  (set-subgoal-terms sg (copy-termlist (subgoal-terms subgoal)))
  (set-subgoal-links sg (subgoal-links subgoal))
  (set-subgoal-copy sg clause-copy)
  (set-subgoal-selected sg #f)
  (set-subgoal-next sg #f)
  (set-subgoal-number sg (subgoal-number subgoal))
  sg))

;(define (make-eq-subgoal s t)
;;; makes an equal-subgoal with termlist (s t)
; (let ((sg (make-vector 7)))
;  (set-subgoal-sign sg '-)
;  (set-subgoal-pred sg 'equal)
;  (set-subgoal-terms sg (list s t))
;  (set-subgoal-links sg equality-links)
;  (set-subgoal-copy sg clause-copy)

(define (make-subgoals-copy subgoals generic length cp)
;; constructs a variable-renamed copy of the <generic> subgoals of a clause.
;; subgoals is the vector of length k+1 (with k subgoals in the clause)
;; which is destructively modified in the procedure. 
;; The procedure has no output.
;; Since the trail is used for keeping the variable consistency, 
;; - before each call to make-subgoal-copy, one has to execute (trail-stop)
;; - and afterwards one has to execute (unbind-vars)
 (if (>= length 0)
     (begin (vector-set! subgoals length
                         (make-subgoal-copy (vector-ref generic length) cp))
            (make-subgoals-copy subgoals generic (- length 1) cp))))

(define (make-clause-copy generic length variables clause constraints)
;; computes a new variable-renamed <copy> of a clause.
;; The trail is used during the procedure, but restored afterwards 
;; Except for possible trail extensions, the procedure is nondestructive 
 (let ((cp (make-vector 9)) (subgoals (make-vector (+ length 1) #f)))
  (trail-stop)
;  (set-copy-variables cp '())
  (make-subgoals-copy subgoals generic (- length 1) cp)
  (set-copy-subgoals cp subgoals)
  (set-copy-clause cp clause)
  (set-copy-constraints cp constraints)
  (unbind-vars)
  cp))

;(define (make-eq-subgoals copy s t index)
; (if (>= index 0)
;     (begin
;       ())))

;;;END *********************************************************************
;;;*************************************************************************
;; ---------------------
;; UNIFICATION PROCEDURE
;; ---------------------

(define (ref term)
;; if term is a <bound_variable>, its binding is returned, 
;; otherwise term is returned
 (if (or (non-variable term) (free-variable term))
     term
     (ref (car term))))

(define (unify-lists k l)
;; unifies termlists k and l.
;; If unification succeeds, #t is returned, else #f.
;; Side effect: termlists and trail destructively modified.
;; Before every unification, one has to call (trail-stop) in order
;; to mark the trail position for the stored substitution
 (if (null? k)
     #t
     (if (unify-terms (ref (car k)) (ref (car l)))
	 (unify-lists (cdr k) (cdr l))
	 #f)))

(define (unify-terms s t)          
;; unifies terms s and t. 
;; If unification succeeds, #t is returned, else #f.
;; Side effect: terms and trail destructively modified.
;; Before every unification, one has to call (trail-stop) in order
;; to mark the trail position for the stored substitution
 (if (eq? s t)
     #t
     (if (variable s)
         (if (variable t) 
             (bind s t)
	     (if (occurs-list s (cdr t))
                 (unbind s t)
  	         (bind s t)))
         (if (variable t)
             (if (occurs-list t (cdr s))
	         (unbind t s)
                 (bind t s))
             (if (eq? (car s) (car t))
	         (unify-lists (cdr s) (cdr t))
                 (unbind s t))))))

(define (occurs-list var termlist)              
;; returns #t if var occurs in termlist, and #f otherwise
 (if (null? termlist)
     #f
     (if (occurs-term var (ref (car termlist)))
         #t
         (occurs-list var (cdr termlist)))))

(define (occurs-term var term)       
;; returns #t if var occurs in term, and #f otherwise
 (if (eq? var term) 
     #t
     (and (non-variable term) (occurs-list var (cdr term)))))

;;;END *********************************************************************

;;;*************************************************************************
;; -------------------------------
;; OPTIMIZED UNIFICATION PROCEDURE
;; -------------------------------

;; For extension steps it is known that initially the two termlists are
;; variable disjoint. Under these conditions the unification can be 
;; optimized such that certain occurs-checks can be omitted, as follows.
;; We assume that for bindings always the first argument position (the k 
;; or s-position, i.e. the entry literal) is preferred; also, we suppose 
;; that no ref-cells are in the terms at the first argument position. 
;; 
;; Initially, the fully optimized procedures extension-unify-* are called.
;; As long as no binding is performed, we remain in this routines.
;;     THE REST IS TO BE CORRECTED ...
;; If the first binding is performed, say, for variable x, then
;; 1. if x is on the second argument position (l or t-position, i.e. from 
;;    the calling subgoal), then we move to standard unification for the 
;;    rest of the unification;
;; 2. otherwise (x must be a free variable) we move to the semi-optimized
;;    procedures semi-extension-* which work as follows:
;;    only free variables at the first argument positions may be bound 
;;    without performing occurs-checks; whenever a variable at the second 
;;    argument position is bound, we move to standard unification.

(define (extension-unify-lists k l)
;; like unify-lists, but fully optimized for extension steps;
;; it is assumed that no binding has been performed yet
 (if 
  (null? k)
  #t
  (if (extension-unify-terms (car k) (ref (car l)))
      (cond ((eq? optimized-unification 0) 
             (extension-unify-lists (cdr k) (cdr l)))
            ((eq? optimized-unification 1) 
             (semi-extension-unify-lists (cdr k) (cdr l)))
	    (else (unify-lists (cdr k) (cdr l))))
      #f)))

(define (extension-unify-terms s t)          
;; like unify-terms, but fully optimized for extension steps;
;; it is assumed that no binding has been performed yet
 (if (variable s)
     (begin (set! optimized-unification 1) (bind s t))
     (if (variable t)
         (begin (set! optimized-unification #f) (bind t s))
         (if (eq? (car s) (car t))
	     (extension-unify-lists (cdr s) (cdr t))
             (unbind s t)))))

(define (semi-extension-unify-lists k l)
;; like unify-lists, but semi-optimized for extension steps;
;; it is assumed that only bindings for variables at the first argument
;; positions have been performed yet
 (cond ((null? k) #t)
       ((free-variable (car k)) (bind (car k) (ref (car l)))
        (semi-extension-unify-lists (cdr k) (cdr l)))
       ((variable (car k)) 
        (if (unify-terms (ref (car k)) (ref (car l)))
            (semi-extension-unify-lists (cdr k) (cdr l))
            #f))
       ((semi-extension-unify-terms (car k) (ref (car l)))
        (if (eq? 1 optimized-unification)
            (semi-extension-unify-lists (cdr k) (cdr l))
            (unify-lists (cdr k) (cdr l))))
       (else #f)))

(define (semi-extension-unify-terms s t)          
;; like unify-terms, but semi-optimized for extension steps;
;; it is assumed that only bindings for variables at the first argument
;; positions have been performed yet
 (if (variable t)
     (if (occurs-list t (cdr s))
	 (unbind t s)
         (begin (set! optimized-unification #f) (bind t s)))
     (if (eq? (car s) (car t))
	 (semi-extension-unify-lists (cdr s) (cdr t))
         (unbind s t))))

;;;END**********************************************************************
;;;*************************************************************************
;; ---------------------------------
;; DIFFERENCE UNIFICATION PROCEDURES
;; ---------------------------------
;; This unification procedure always succeeds. If two terms are unifiable,
;; the procedure behaves like ordinary unification. If, however, a failure
;; point in the standard unification would be reached, this procedure stores 
;; the non-unifiable pair of terms and proceeds.

(define (difference-unify-lists k l)
;; difference-unifies termlists k and l.
;; Side effect: termlists and trail destructively modified.
;; Before every unification, one has to call (trail-stop) in order
;; to mark the trail position for the stored substitution
 (if (pair? k)
     (begin (difference-unify-terms (ref (car k)) (ref (car l)))
	    (difference-unify-lists (cdr k) (cdr l)))))

(define (difference-unify-terms s t)          
;; difference-unifies terms s and t. 
;; Side effect: terms and trail destructively modified.
;; Before every unification, one has to call (trail-stop) in order
;; to mark the trail position for the stored substitution
 (if (eq? s t)
     #t
     (if (variable s)
         (if (variable t) 
             (bind s t)
	     (if (occurs-list s (cdr t))
                 (store-difference s t)
  	         (bind s t)))
         (if (variable t)
             (if (occurs-list t (cdr s))
                 (store-difference t s)
                 (bind t s))
             (if (eq? (car s) (car t))
	         (difference-unify-lists (cdr s) (cdr t))
                 (store-difference s t))))))

;;;END**********************************************************************
;;;*************************************************************************
;; -----------------------------
;; TRAIL MODIFICATION PROCEDURES

(define (unbind-vars)
;; makes free all variables back to the previous trailstop
 (if (trail-content)
     (begin (set-car! (caar trail) '()) (trail-previous) (unbind-vars))
     (begin (trail-previous) #f)))

(define (unbind-collect-vars variables)
;; makes free and collects all variables back to the previous trailstop
 (if (trail-content)
     (begin (set-car! (caar trail) '()) 
	    (set! variables (cons (caar trail) variables))
	    (trail-previous)
	    (unbind-collect-vars variables))
     (begin (trail-previous) variables)))

(define (last-binding-variables)
 (let ((old-trail trail) variables)
      (set! variables (last-binding-vars '()))
      (set! trail old-trail)
      variables))

(define (last-binding-vars variables)
 (if (trail-content)
     (begin (set! variables (cons (caar trail) variables))
	    (trail-previous)
	    (last-binding-vars variables))
     variables))
;;;*************************************************************************
;; ---------------------------
;; CONNECTION GRAPH PROCEDURES
;; ---------------------------

(define (make-connection-graph clauses-out clauses-in)
;; computes the links for every literal occurrence in the clauses
;; Procedure is destructive in that it modifies the subgoal-links
;; position of each subgoal vector
 (if (pair? clauses-out)
     (begin (make-cg-clause (clause-generic (car clauses-out)) 
                            0 clauses-in)
            (make-connection-graph (cdr clauses-out) clauses-in))))

(define (make-cg-clause generic offset clauses-in)
 (if (vector-ref generic offset)
     (begin 
      (make-cg-subgoal (vector-ref generic offset) clauses-in)
      (set-subgoal-links (vector-ref generic offset)
                   (reverse (subgoal-links (vector-ref generic offset))))
      (if (> nr-non-horn 0)
          (set-subgoal-links (vector-ref generic offset)
               (cons 'reduction (subgoal-links (vector-ref generic offset)))))
      (set-subgoal-links      ; updates links in first clause copy
                   (subgoals-offset-copy generic offset)
                   (subgoal-links (vector-ref generic offset)))
      (make-cg-clause generic (+ 1 offset) clauses-in))))

(define (make-cg-subgoal subgoal clauses-in)
 (if (pair? clauses-in)
     (begin (make-cg-sg-clause subgoal
                  (copy-subgoals (car (clause-copies (car clauses-in))))
		  0 (clause-length (car clauses-in)))
            (make-cg-subgoal subgoal (cdr clauses-in)))))

(define (make-cg-sg-clause subgoal subgoals offset length)
 (if (< offset length)
     (begin (make-cg-sg-sg subgoal (vector-ref subgoals offset) offset)
            (make-cg-sg-clause subgoal subgoals (+ offset 1) length))))

(define (make-cg-sg-sg subgoal other-subgoal offset)
 (if (and (eq? (subgoal-pred subgoal) (subgoal-pred other-subgoal))
          (not (eq? (subgoal-sign subgoal) (subgoal-sign other-subgoal)))
          (or equality
              (unify-lists-stop (subgoal-terms other-subgoal) 
	                        (subgoal-terms subgoal))))
     (begin (or equality (unbind-vars))
            (set-subgoal-links subgoal 
	         (cons (cons (copy-clause (subgoal-copy other-subgoal)) offset)
                       (subgoal-links subgoal))))))

(define (make-eq-links clauses-in)
 (if (pair? clauses-in)
     (begin (make-eq-links-clause 
                  (copy-subgoals (car (clause-copies (car clauses-in))))
		  0 (clause-length (car clauses-in)))
            (make-eq-links (cdr clauses-in)))
     (begin (set! eq-links (reverse eq-links))
            (if (> nr-non-horn 0)
                (set! eq-links (cons 'reduction eq-links))))))

(define (make-eq-links-clause subgoals offset length)
 (if (< offset length)
     (begin (make-eq-links-sg (vector-ref subgoals offset) offset)
            (make-eq-links-clause subgoals (+ offset 1) length))))

(define (make-eq-links-sg subgoal offset)
 (if (and (eq? 'equal (subgoal-pred subgoal)) (eq? '+ (subgoal-sign subgoal)))
     (set! eq-links 
        (cons (cons (copy-clause (subgoal-copy subgoal)) offset) eq-links))))

;(define (insert-reduction clauses)
; (if (pair? clauses)
;     (begin (insert-reduction-literals (clause-generic (car clauses))
;                                       (- (clause-length (car clause)) 1)
;            (insert-reduction (cdr clauses)))))
;
;(define (insert-reduction-literals subgoal-vector index)
; (if 
;  (>= index 0)
;  (begin 
; (if (> (clause-length (caar (subgoal-links (vector-ref subgoal-vector index))
;   (set-subgoal-links (vector-ref subgoal-vector index)
;   (insert-reduction-links (subgoal-links (vector-ref subgoal-vector index))))
;   (insert-reduction-literals subgoal-vector (- index 1)))))
;
;(define (insert-reduction-links links new-links)
; (if (pair? links)
;     (if (= 1 (clause-length (car links))) 
;         (insert-reduction-links (cdr links))
;         (set-cdr! links 
;;;END *********************************************************************
;;;*************************************************************************
;; --------------------------------------
;; MODEL ELIMINATION PROOF PROCEDURE
;; --------------------------------------

(define (solve sg links depth inference)
 (if 
  (not (null? links))
  (if (and sgindex (find-matching-lemma sg)
;	   (set! i (+ 1 i)) #f
	   )
      (let ((old-global-subgoals global-subgoals) (new-inf inference)
            new-sg new-links new-depth)
        (set! successful-inferences (+ successful-inferences 1))
	(set! successful-matchings (+ successful-matchings 1))
	(set! nr-matchings (+ 1 nr-matchings))
	(next-subgoal-call-fast)
	(set! nr-matchings (- nr-matchings 1)))
      (begin 
	(if (eq? (car links) 'reduction)
	    (reduction-steps sg (copy-path (subgoal-copy sg)) inference)
	    (extension sg (caar links) 
		       (vector-ref 
 			   (copy-subgoals (car (clause-copies (caar links))))
			   (cdar links))
		       depth inference))
	(if subgoal-alternation 
	    (alternate-subgoals sg links depth inference)
	    (solve sg (cdr links) depth inference))))))

(define (reduction-steps sg predsg inference)
 (if 
  predsg
  (begin
   (if 
    (and (not (eq? (subgoal-sign sg) (subgoal-sign predsg)))
         (eq? (subgoal-pred sg) (subgoal-pred predsg))
         (or (not matings-pruning) ; begin of matings pruning
             (> (subgoal-clause-number sg)
                (subgoal-clause-number (subgoal-next predsg)))
             (and (= (subgoal-clause-number sg)
                     (subgoal-clause-number (subgoal-next predsg)))
                  (> (subgoal-number sg)
                     (subgoal-number (subgoal-next predsg))))) ; end mat. pr.
         (or (and (unify-lists-stop (subgoal-terms sg) (subgoal-terms predsg))
	          (or (not constraints)
		      (begin (v-trail-stop) ; backtracking marker
			(set! actual-c-stack c-stack)
			(if (propagate-substitution) ; constr. not viol.
			    #t ; below: shallow constraint and unfication
			    (begin (set! c-stack actual-c-stack) ; backtracking
				(undo-v-trail) (unbind-vars) 
				(set! constraint-fails (+ constraint-fails 1))
				#f))))) ; next line ends shallow backtracking
             (begin (set! failed-inferences (+ failed-inferences 1)) #f)))
    (let ((old-global-subgoals global-subgoals) (new-inf inference)
          (old-c-stack actual-c-stack) new-sg new-links new-depth)
     (set! successful-inferences (+ successful-inferences 1))
     (set! successful-reduction-steps (+ successful-reduction-steps 1))
     (set! nr-reduction-steps (+ nr-reduction-steps 1))
;     (next-subgoal-call new-inf new-sg new-links new-depth old-global-subgoals)
     (next-subgoal-call-fast)
     ;; deep backtracking:
     (set! nr-reduction-steps (- nr-reduction-steps 1))
     (if constraints (begin (set! c-stack old-c-stack) (undo-v-trail)))
     (unbind-vars)))
   (reduction-steps sg (copy-path (subgoal-copy predsg)) inference))))
     
(define (extension sg clause entry depth inference)
 (if (and (check-bounds clause depth)
          (or (not signature-bound)
              (begin (set! old-free-sig free-sig)
                     (set! old-sig-trail-position sig-trail-position)
                     (if (sig-bound (clause-signature clause))
                         #t
                         (begin (set! free-sig old-free-sig) 
                                (set! bound-fail #t) #f))))
          (or (begin (set! actual-depth depth)
                     (set! actual-copy (car (clause-copies clause)))
		     (not sg)) ; top clause entered: no unification/regularity
              (and (extension-unify-lists-stop (subgoal-terms entry) 
					       (subgoal-terms sg))
		   (or (not constraints)
		       (begin (v-trail-stop) ; backtracking marker
			(set! actual-c-stack c-stack)
			(if (and (propagate-substitution) ; constr. not viol.
				 (regularity entry (copy-subgoals actual-copy) 
					     sg (clause-length clause)))
			    #t ; below: shallow constraint and unfication
			    (begin (set! c-stack actual-c-stack) ; backtracking
				(undo-v-trail) (unbind-vars) 
				(set! constraint-fails (+ constraint-fails 1))
				#f)))))
	      (begin (if signature-bound ; shallow signature backtracking
			 (begin (untrail-sig old-sig-trail-position)
				(set! free-sig old-free-sig)))
		     #f)))
     (let ((copy actual-copy) (old-global-subgoals global-subgoals)
	   (old-copies (clause-copies clause)) (new-inf inference)
           (old-eq-trail eq-trail-before) 
           new-sg new-links new-depth
           (old-local-free-sig old-free-sig)
           (old-local-sig-trail-position old-sig-trail-position)
           (old-c-stack actual-c-stack))
      (set! successful-inferences (+ successful-inferences 1))
;      (if sg (begin (display (clause-number clause)) (display ", ")))
      (set-copy-path copy sg) (new-clause-copy clause)
      (set-clause-copies clause (cdr (clause-copies clause)))
;      (set-copy-eq-counter copy difference-counter)
      (set! new-inf (+ difference-counter new-inf (clause-length clause)))
      (if sg (begin (set! new-inf (- new-inf 1))
                    (set-subgoal-next sg entry)))
      (set! inf new-inf)
      (if (or (not sg) (> (clause-length clause) 1) (> difference-counter 0))
          (begin 
            (set-copy-depth copy (- depth 1))
            (set-copy-inference copy new-inf)
            (if weighted-depth
                (set-copy-free-depth100 copy 
                   (- (+ (* 100 (copy-depth copy)) wd1)
                      (* wd2 (+ (clause-length clause)
				difference-counter -1)))))
	    (if (or (not sg) (> (clause-length clause) 1))
	        (begin (vector-set! (copy-subgoals copy) 
			            (clause-length clause) global-subgoals)
                       (set! global-subgoals (copy-subgoals copy))))
            (if (> difference-counter 0)
                (begin (set-subgoal-other-subgoals (car eq-trail) 
                                                   global-subgoals)
                       (set! global-subgoals (copy-eq-trail copy)))
                (set-copy-eq-trail copy #f))
            (if sg 
                (set-subgoal-selected entry #t)
	        (set! all-global-subgoals global-subgoals))))
;      (next-subgoal-call new-inf new-sg new-links new-depth 
;			 old-global-subgoals)
      (next-subgoal-call-fast)
      ;; deep backtracking:
      (if sg 
          (begin (set-subgoal-selected entry #f)
                 (if constraints
                     (begin (set! c-stack old-c-stack) (undo-v-trail)))
                 (set! eq-trail old-eq-trail)
                 (unbind-vars)))
      (if signature-bound (begin (untrail-sig old-local-sig-trail-position)
				 (set! free-sig old-local-free-sig)))
      (set-clause-copies clause old-copies))
      ;; end of shallow backtracking:
      (set! failed-inferences (+ failed-inferences 1))))
 
;;;*************************************************************************
;; -----------------
;; SUBGOAL SELECTION
;; -----------------

(define (select-subgoal subgoals index max-index)
 (if 
  (= index max-index)
   (if (and subgoal-reordering current-subgoal)
       (if (and global-reordering 
                (vector? (vector-ref subgoals max-index))
                (> (subgoal-depth current-subgoal) 0))
           (select-subgoal (vector-ref subgoals max-index) 0
                (- (vector-length (vector-ref subgoals max-index)) 1))
           current-subgoal)
       (begin 
	(set! global-subgoals (vector-ref subgoals max-index))
	(cond ((vector? global-subgoals)
	       (select-subgoal global-subgoals 0 
                             (- (vector-length global-subgoals) 1)))
              ((pair? global-subgoals) (select-eq-subgoal global-subgoals))
	      (else (display-inferences depth)
	            (newline) (display "     ***** PROOF WITH ") 
		    (display inf) (display " INFERENCES ")
                    (if (> nr-reduction-steps 0) 
                        (begin (display "(") (display nr-reduction-steps) 
                               (display " RED.) ")))
		    (display "FOUND *****") (newline) (newline)
		    (display "Total successful/failed inferences")
;		    (if constraints (display "(constraints)"))
		    (display ": ") 
		    (display total-successful-inferences)
                    (if (> successful-reduction-steps 0)
                        (begin (display " (RED: ") 
		               (display successful-reduction-steps)
                               (display ")")))
                    (display "/")
		    (display total-failed-inferences) 
;		    (if constraints 
;			(begin (display " (") 
;			       (display total-constraint-fails) (display ")")))
		    (newline)
                    (if proof-output 
                        (begin (newline) (newline) (display "USED CLAUSES")
                           (newline) (display "------------") (newline)
                           (display-proof-clauses (formula-clauses formula))))
		    (exit 0) #f))))
  (if (subgoal-selected (vector-ref subgoals index))
      (select-subgoal subgoals (+ index 1) max-index)
      (if 
        subgoal-reordering
	(begin (compare-subgoals (vector-ref subgoals index))
               (select-subgoal subgoals (+ index 1) max-index))
        (vector-ref subgoals index)))))

(define (select-eq-subgoal eq-trail)
 (if (subgoal-selected (car eq-trail))
     (if (subgoal-other-subgoals (car eq-trail)) ;; end of current eq-subgoals
         (if (not current-subgoal) ;; no current eq-subgoal
             (begin 
               (set! global-subgoals (subgoal-other-subgoals (car eq-trail)))
               (if (pair? global-subgoals) 
                   (select-eq-subgoal global-subgoals)
                   (select-subgoal global-subgoals 0 
                                   (- (vector-length global-subgoals) 1))))
             current-subgoal)
         (select-eq-subgoal (cdr eq-trail)))
     (if subgoal-reordering    
         (if current-subgoal 
             (begin (compare-subgoals (car eq-trail))
                    (select-eq-subgoal (cdr eq-trail)))
             (begin (set! current-subgoal (car eq-trail))
                    (select-eq-subgoal (cdr eq-trail))))
         (car eq-trail))))

(define (compare-subgoals subgoal)
 (if current-subgoal
     (if current-complexity 
	 (begin (trail-stop)
;	        (set! new-complexity (termlist-v-complexity 
;		                                 (subgoal-terms subgoal)))
	        (set! new-complexity (subgoal-instantiation subgoal))
		(set! new-vars (unbind-bindings))
		(if (subgoal-conditions subgoal)
		    (begin (set! current-subgoal subgoal)
		           (set! current-vars new-vars)
                           (set! current-complexity new-complexity))))
         (begin (trail-stop)
;		(set! current-complexity (termlist-v-complexity 
;		                           (subgoal-terms current-subgoal)))
	        (set! current-complexity 
                      (subgoal-instantiation current-subgoal))
                (set! current-vars (unbind-bindings))
	        (trail-stop)
;	        (set! new-complexity (termlist-v-complexity 
;		                           (subgoal-terms subgoal)))
	        (set! new-complexity (subgoal-instantiation subgoal))
		(set! new-vars (unbind-bindings))
		(if (subgoal-conditions subgoal)
		    (begin (set! current-subgoal subgoal)
		           (set! current-vars new-vars)
                           (set! current-complexity new-complexity)))))
    (set! current-subgoal subgoal)))

(define (unbind-bindings)
;; makes free all variables back to the previous trailstop, outputs number
;; of modified bindings 
 (if (trail-content)
     (begin (set-car! (caar trail) '()) 
            (trail-previous) (+ 1 (unbind-bindings)))
     (begin (trail-previous) 0)))

(define (ref-int int-term)
;; if int-term is a <bound_variable>, its binding is returned, 
;; otherwise int-term is returned
 (if (or (integer? int-term) (free-variable int-term) 
	 (vector? (car int-term)))
     int-term
     (ref-int (car int-term))))

(define (ref-ext ext-term)
;; if term is a <bound_variable>, its binding is returned, 
;; otherwise term is returned
 (if (or (not ext-term) (non-variable ext-term) (free-variable ext-term))
     ext-term
     (ref-ext (car ext-term))))

;;(define (ref-occ occ-ground-term)
;;;; an occ-ground-term is a bound variable, a term or a nonnegative integer
;;;; if term is a <bound_variable>, its binding is returned, 
;;;; otherwise occ-ground-term is returned
;; (if (or (non-variable occ-ground-term) (free-variable ext-term))
;;     ext-term
;;     (ref-ext1 (car ext-term))))

(define (termlist-var-occ terms)
; counts number of variables in term by setting the variable to
; its current frequency and incrementing this value;
; uses the trail: therefore, before calling main function, say (trail-stop)
 (if (pair? terms)
     (begin (term-var-occ (car terms))
;            (newline) (display "1 :") (set! l1 terms) (repl)
            (termlist-var-occ (cdr terms)))))

(define (term-var-occ term)
; instantiates free variables to 0 and variables with integer bindings n
; to n+1; auxiliary to procedure termlist-var-occ 
 (if (free-variable term) 
     (bind term 0)
     (if (integer? (car term))
         (set-car! term (+ (car term) 1))
         (if (vector? (car term))
             (termlist-var-occ (cdr term))
             (term-var-occ (car term))))))

(define (term-instantiation1 term)
; computes instantiation value of an extended term t as follows:
;           |  ( occ(t) - 1 ) / occ(t)    if t is a variable (occurrence)
; t-i(t) =  {  1                          if t is a constant 
;           |  1/2 + sum(t-i(a_i)/2*n)    if t = f(a_1,...,a_n)
;
 (cond ((integer? (car term))
        (quotient (* 1000 (car term)) (+ (car term) 1)))
       ((constant term) 1000)
       ((vector? (car term))
        (+ 500 (quotient (termlist-instantiation1 (cdr term)) 
                              (* 2 (functor-arity (car term))))))
       (else (term-instantiation1 (car term)))))

(define (termlist-instantiation1 termlist)
 (if (null? (cdr termlist))
     (term-instantiation1 (car termlist))
     (+ (term-instantiation1 (car termlist))
        (termlist-instantiation1 (cdr termlist)))))

(define (termlist-v-complexity terms)
; computes number of symbols in terms - number of distinct variables
; uses the trail: therefore, before calling main function, say (trail-stop)
;                            and afterwards say: (unbind-bindings))
 (if (null? terms)
     0
     (+ (term-v-complexity (ref-ext (car terms))) 
        (termlist-v-complexity (cdr terms)))))

(define (term-v-complexity term)
; computes number of symbols in term - number of distinct variables
; uses the trail: therefore, before calling main function, say (trail-stop)
;                            and afterwards say: (unbind-bindings)
 (cond ((or (not term) (constant term)) 1)
       ((variable term) (bind term #f) 0)
       (else (+ 1 (termlist-v-complexity (cdr term))))))

(define (term-instantiation term)
; computes instantiation value of a term as follows:
;           |  0                          if t is a variable
; t-i(t) =  {  1                          if t is a constant 
;           |  1/2 + sum(t-i(a_i)/2*n)    if t = f(a_1,...,a_n)
;
 (cond ((not term) 0)
       ((variable term) (bind term #f) 0)
       ((constant term) 1000)
       (else (+ 500 (quotient (termlist-instantiation (cdr term)) 
                              (* 2 (functor-arity (car term))))))))

(define (termlist-instantiation termlist)
 (if (null? (cdr termlist))
     (term-instantiation (ref-ext (car termlist)))
     (+ (term-instantiation (ref-ext (car termlist)))
        (termlist-instantiation (cdr termlist)))))

(define (compute-weighted-depth depth free-depth100 entry-inf inf) 
; first line must remain, otherwise segmentation fault 
; (probably because of compiler optimization)
    (quotient (+ 50 free-depth100
                 (quotient (* 100 (- (* depth 100) free-depth100))
                           (+ 100 (* (- inf entry-inf) wd3))))
              100))

;;;*************************************************************************
;; ------------------------
;; COMPUTE AND CHECK BOUNDS
;; ------------------------

(define (free-signature-weights sigs)
 (if (pair? sigs)
     (begin (set-signature-weight (car sigs) 0)
            (free-signature-weights (cdr sigs)))))
 
(define (update-pred-numbers preds)
 (if (pair? preds)
     (begin (set-predsym-number (car preds) 
                         (+ (predsym-number (car preds)) nr-functors))
            (update-pred-numbers (cdr preds)))))
 
(define (sig-bound sigs)
;; the function updates free-sig and the signature vector sig-vector: 
;; free-sig is decremented with the sum of the signature weights of the 
;; new symbols in sigs and the respective positions in the sig-vector are
;; incremented with the signature weights; the new positions in the vector 
;; are remembered in the sig-trail;
;; sig-bound returns #f if the new value of free-sig is < 0
 (if (pair? sigs)
     (if (not (vector-ref sig-vector (signature-number (car sigs))))
         (begin (set! free-sig (- free-sig (signature-weight (car sigs))))
           (if (and (>= free-sig 0) (sig-bound (cdr sigs)))
               (begin (vector-set! sig-vector (signature-number (car sigs))
                                              (signature-weight (car sigs)))
                 (vector-set! sig-trail sig-trail-position 
                                        (signature-number (car sigs)))
                 (set! sig-trail-position (+ sig-trail-position 1)))
               #f))
         (sig-bound (cdr sigs)))
     #t))

(define (untrail-sig old-sig-trail-position)
 (if (> sig-trail-position old-sig-trail-position)
     (begin 
       (set! sig-trail-position (- sig-trail-position 1))
       (vector-set! sig-vector (vector-ref sig-trail sig-trail-position) #f)
       (untrail-sig old-sig-trail-position))))

(define (term-variables termlist)
; extracts set of variables occurring in termlist
; nondestructive, but internally misuses the trail
 (trail-stop)
 (termlist-var-occ termlist)
 (unbind-collect-vars '()))

(define (subgoals-termlists-appended subgoals index)
 (if (< (vector-length subgoals) index)
     (append (subgoal-terms (vector-ref subgoals index))
	     (subgoals-termlists-appended subgoals (+ index 1)))
     '()))

(define (next-subgoal-call-old new-inf new-sg new-links new-depth
			   old-global-subgoals)
 (set! current-subgoal #f) (set! current-complexity #f) (set! new-sg #f)
;  (if (pair? global-subgoals) ;; global-subgoals is an eq-trail
;      (set! new-sg (select-eq-subgoal global-subgoals))
      (set! new-sg (select-subgoal global-subgoals 0 
                                   (- (vector-length global-subgoals) 1)))
;)
  (if 
   (and new-sg (equal-predecessor-check new-sg))
   (begin 
    (set! new-links (subgoal-links new-sg))
    (if weighted-depth
        (set! new-depth (compute-weighted-depth
                                 (subgoal-depth new-sg)
                                 (copy-free-depth100 (subgoal-copy new-sg))
                                 (copy-inference (subgoal-copy new-sg))
                                 new-inf))
        (set! new-depth (subgoal-depth new-sg)))
    (set! actual-depth new-depth)
    (set-subgoal-selected new-sg #t)
    (solve new-sg new-links new-depth new-inf)
    (set-subgoal-selected new-sg #f)
    (set-subgoal-links new-sg new-links)))
  (set! global-subgoals old-global-subgoals))
;;;***********************************************************************
;; ----------------------
;; EQUAL PREDECESSOR FAIL
;; ----------------------

(define (no-equal-predecessor sg sg-terms predsg)
 (if 
  (and predsg (not (free-variable-in sg-terms)))
  (if (and (eq? (subgoal-sign sg) (subgoal-sign predsg))
           (eq? (subgoal-pred sg) (subgoal-pred predsg))
           (ref-equals sg-terms (subgoal-terms predsg)))
      #f
      (no-equal-predecessor sg sg-terms (subgoal-predecessor predsg)))
  #t))

(define (ref-equals k l)
;; checks whether termlists k and l are equal 
 (if (null? k)
     #t
     (if (ref-equal (ref (car k)) (ref (car l)))
	 (ref-equals (cdr k) (cdr l))
	 #f)))

(define (ref-equal s t)          
;; checks whether terms s and t are equal 
 (if (eq? s t)
     #t
     (if (or (variable s) (variable t))
         #f
         (and (eq? (car s) (car t)) 
              (ref-equals (cdr s) (cdr t))))))

(define (free-variable-in terms)
 (if (pair? terms)
     (if (free-variable (car terms))
         #t
         (free-variable-in (cdr terms)))
     #f))
 
;;;***********************************************************************
;; ----------------------
;; REGULARITY CONSTRAINTS
;; ----------------------

(define (make-regularity-constraints-clause entry subgoals predsg index
					    max-index)
; generates regularity-constraints for all subgoals/literals of a clause
; output: #t if none of the generated constraints is violated, otherwise #f
; side effects on the constraints are stored on the value trail
;  (print "make-regularity-constraints-clause entered:")
;  (display-subgoal (vector-ref subgoals index))
;  (display-subgoal entry)
;  (newline)
(if (= index max-index) ; constraints for all subgoals in clause generated
     #t
     (if (or (and (eq? entry (vector-ref subgoals index)) ; head literal
		  regularity-only-for-subgoals)
	     (let ((result 
	     (make-regularity-constraints (vector-ref subgoals index) 
	                  (subgoal-terms (vector-ref subgoals index)) predsg))
		   )
	;         (display "RESULT: ") (display result) (newline)
	       result))
	 (make-regularity-constraints-clause entry subgoals predsg (+ 1 index)
					     max-index)
	 #f))) ; some regularity constraint is violated

(define (make-regularity-constraints sg sg-terms predsg)
;  (print "make-regularity-constraints entered:")
;  (set! l1 sg) (set! l2 predsg)
;  (display-subgoal sg)
;  (if predsg (display-subgoal predsg))
;  (repl)
(if predsg 
     (if (and (eq? (subgoal-sign sg) (subgoal-sign predsg))
	      (eq? (subgoal-pred sg) (subgoal-pred predsg)))
	 (and (make-disequation-constraint sg-terms (subgoal-terms predsg))
	      (make-regularity-constraints sg sg-terms
					   (subgoal-predecessor predsg)))
	 #t)
     #t))

(define (make-unit-subsumption-constraint subgoal unit-terms constraints)
; generates subsumption constraint for a subgoal whose subgoal list unifies
; (with unifier sigma) with unit-terms. 
; Side effect: integrates constraint with variables not in subgoal made
; universal into the constraints of the variables of the subgoal.
; Output: #f if sigma is the empty substitution, otherwise #t
 (if (unify-lists-stop (subgoal-terms subgoal) unit-terms)
     (if (trail-content) ; constraint has length > 0
	 (begin (set! constraints   ; new header is made
		      (cons (vector 0 #f #f) constraints))
		(extract-constraint-static 1 (car constraints) 
					   constraints (cdr constraints))
		(trail-previous) #t)
	 (begin (trail-previous) #f)) ; constraint is immediately violated
     #t)) ; term lists not unifiable: no constraint generated

(define (extract-constraint-static length c-header constraints previous)
; makes free all variables back to the previous trailstop
; and pushes the bindings as subconstraints on the list constraints
; (caar trail) is current variable, (car constraints) is new subconstraint
; previous is the previous subconstraint of the respective variable
 (set! constraints
       (cons (vector (ref (var-binding (caar trail))) c-header previous)
	     constraints))
 (set-car! (caar trail) '()) ; var. binding from protocol unification undone 
 (trail-previous) ; moves to previous trail-point 
 (if (trail-content) ; next binding to be processed
     (extract-constraint-static (+ length 1) c-header constraints 
				(cdr constraints))
     ; below: all bindings processed; now constraint length inserted in header
     (vector-set! c-header 0 (+ (vector-ref c-header 0) length))
     )); General information.
; --------------------
; This is the code for integrating disequation constraints into Scheme-Setheo.
; General changes to the prover are the following.
;
; 1) Generation of structural constraints.
; After every extension step, structural constraints (regularity) is generated
; and pushed on the constraint stack, and the constraint lists of the
; respective variables are modified. The basic procedure for normalizing and
; integrating a constraint is make-disequation-constraint. It uses the
; recursive procedure extract-constraint as subroutine.
;
; 2) Propagation of substitutions to constraints.
; After every successful inference step, the last computed substitution is
; propagated to the constraints. The main procedure for doing this is
; propagate-substitution. Again the procedure extract-constraint is used as
; subroutine. 
; 
; 3) Backtracking.
; The changes performed on the constraint data are stored in the value trail,
; which is organized similar to the trail (for variable bindings). 
; Before every sequence of constraint modifications associated with an 
; inference step, a v-trail-stop is set. When backtracking an inference step,
; the changes documented on the value trail have to be undone (from the
; current top of trail (the global variable v-trail) back to the previous
; v-trail-stop. The respective procedure is undo-v-trail. 
;
; Important note: if a constraint violation occurs during the generation or
; propagation of constraints, then the unwinding of the value stack is not
; done implicitly in those procedures, but most be done explicitly afterwards.
; Furthermore, in order to permit reuse of the constraint stack, the global
; variable c-stack has to be set back to the old top of the constraint stack.
; Both operations are done in the procedures extension and reduction-steps.
;
; Relevant data structures:
; -------------------------
; VARIABLE is ( VALUE CONSTRAINTS . NAME )
;
; CONSTRAINT-STACK is list of SUBCONSTRAINTS or HEADERS
; global variable c-stack-root is entire CONSTRAINT-STACK
; global variable c-stack is actual top of CONSTRAINT-STACK
; SUBCONSTRAINT is #( BINDING HEADER NEXT-SUBCONSTRAINT )
; HEADER is #( LENGTH #t/#f DUMMY )
;
; VALUE-TRAIL is list of VALUE-TRAILPOINTS
; VALUE-TRAILPOINT is #( STRUCT POS OLD-VALUE PREVIOUS-VALUE-TRAILPOINT ) 
; POS is car, cdr (for cons cells), or a natural number (the vector index)
; global variable v-trail-root is entire VALUE-TRAIL
; global variable v-trail is actual top of VALUE-TRAIL

;;;*************************************************************************
;; -------------------------------
;; GENERATE DISEQUATION CONSTRAINT
;; -------------------------------

(define (make-disequation-constraint l1 l2)
; integrates the normalization of the disequation constraint l1 =/= l2
; output: #f if the constraint is already violated, otherwise #t.
; Side effects on the constraints are stored on the value trail.
; Note: the procedure misuses the variable trail, which is restored afterwards
 (if (unify-lists-stop l1 l2) ; protocol unification performed
     (if (trail-content) ; constraint has length > 0
	 (begin (constraint-stack-next) ; new header is made
		(vector-set! (car c-stack) 0 0) ; length in header set to 0
		(vector-set! (car c-stack) 1 #f) ; constraint is not true
		(extract-constraint 1 (car c-stack))
		(trail-previous) #t)
	 (begin (trail-previous) #f)) ; constraint is immediately violated
     #t)) ; term lists not unifiable: no constraint generated

(define (extract-constraint length c-header)
;; makes free all variables back to the previous trailstop
;; and pushes the bindings as subconstraints on the constraint stack;
;; relevant changes are documented on the value trail;
; (caar trail) is current variable, (car c-stack) is new subconstraint

 (v-trail-next) ; old value of variable constraint list is stored:
 (vector-set! (car v-trail) 0 (cdr (caar trail))) ; affected cons cell  
 (vector-set! (car v-trail) 1 'car) ; memorizes that car was changed
 (vector-set! (car v-trail) 2 (var-constraints (caar trail))) ; old value

 (constraint-stack-next) ; new subconstraint generated and added:
 (vector-set! (car c-stack) 0 (ref (var-binding (caar trail)))) ; binding
 (vector-set! (car c-stack) 1 c-header) ; pointer to constraint header
 (vector-set! (car c-stack) 2 (var-constraints (caar trail))) ; old constraints
 (set-car! (cdr (caar trail)) (car c-stack)) ; new variable constraint list 
	    
 (set-car! (caar trail) '()) ; var. binding from protocol unification undone
 (trail-previous) ; moves to previous trail-point 
 (if (trail-content) ; next binding to be processed
     (extract-constraint (+ length 1) c-header)
     ; below: all bindings processed; now constraint length inserted in header
     (vector-set! c-header 0 (+ (vector-ref c-header 0) length))
     ))

;;;*************************************************************************
;; -------------------------------------------------
;; PROPAGATE SUBSTITUTION TO DISEQUATION CONSTRAINTS
;; -------------------------------------------------

(define (propagate-substitution)
; applies current substitution, which is documented on the variable trail,
; to the constraints in the constraint stack;
; output: #t/#f if constraints are satisfiable/unsatisfiable.
; Side effects on the constraints are stored on the value trail.
; Note: the procedure misuses the variable trail, which is restored afterwards
 (if (trail-content) ; substitution is nonempty
     (propagate-bindings trail)
     #t)) ; no propagation to be performed

(define (propagate-bindings b-trail)
; (caar b-trail) is current variable
; b-trail is the trail position of the respective binding; the global variable
; trail cannot be used for this purpose, since the trail is misused for
; protocol unifications in subroutines of this procedure
 (if (propagate-binding (var-binding (caar b-trail)) ; current binding does not
			(var-constraints (caar b-trail))) ; violate constraints
     (begin (set! b-trail (trail-previous-loc b-trail)) ; process next binding 
	    (if (trail-content-loc b-trail) ; remaining substitution nonempty
		(propagate-bindings b-trail)
		#t)) ; all bindings propagated and no constraint violated
     #f)) ; some constraint violated

(define (propagate-binding binding subconstraint)
; subconstraint is the top subconstraint of the respective variable, i.e.,
; either #f or otherwise a vector of length 3 with its last element pointing
; to the next subconstraint
 (if subconstraint
     (if (or (is-passiv-constraint? subconstraint)
	     (expand-subconstraint binding subconstraint))
	 (propagate-binding binding (vector-ref subconstraint 2))
	 #f) ; current constraint is violated
     #t)) ; binding has been propagated to all subconstraints of this variable

(define (expand-subconstraint binding subconstraint)
 (if (unify-terms-stop ; protocol unification performed
		       (ref binding)
		       (ref (vector-ref subconstraint 0))
     )		       
     (if (trail-content) ; new constraint has length > 0
	 (begin (decrement-and-trail-header (vector-ref subconstraint 1))
		(extract-constraint 1 (vector-ref subconstraint 1))
		(trail-previous) #t)
	 (begin (trail-previous) ; subconstraint is violated
		(if (= 1 (vector-ref (vector-ref subconstraint 1) 0)) ; entire
		    #f                                ; constraint is violated
		    (begin (decrement-and-trail-header
			    (vector-ref subconstraint 1))
			   #t))))
     ; below: unification failed, hence constraint deactivated
     (begin (deactivate-and-trail-header (vector-ref subconstraint 1))
	    #t))) ; term lists not unifiable: no new constraint generated

(define (decrement-and-trail-header c-header)
 (v-trail-next) ; change is trailed on the value trail
 (vector-set! (car v-trail) 0 c-header) ; affected vector
 (vector-set! (car v-trail) 1 0) ; memorizes that index 0 was changed
 (vector-set! (car v-trail) 2 (vector-ref c-header 0)) ; old value stored;
 (vector-set! c-header 0 (- (vector-ref c-header 0) 1))) ; header length - 1

(define (deactivate-and-trail-header c-header)
 (v-trail-next) ; change is trailed on the value trail
 (vector-set! (car v-trail) 0 c-header) ; affected vector
 (vector-set! (car v-trail) 1 1) ; memorizes that index 1 was changed
 (vector-set! (car v-trail) 2 (vector-ref c-header 1)) ; old value stored
 (vector-set! c-header 1 #t)) ; constraint deactivated

;;;*************************************************************************
;; ------------------------
;; VALUE TRAIL MODIFICATION
;; ------------------------

(define (undo-v-trail)
; undos all modifications stored on the value trail back to the previous
; trail-stop, then moves one step back to the previous v-trail position
 (if (v-trail-content) ; some modification has to be undone
     (begin (cond ((vector? (vector-ref (car v-trail) 0)) ; vector 
		   (vector-set! (vector-ref (car v-trail) 0)
				(vector-ref (car v-trail) 1) ; index
				(vector-ref (car v-trail) 2))) ; old value
		  ((eq? 'car (vector-ref (car v-trail) 1)) ; car of cell
		   (set-car! (vector-ref (car v-trail) 0) 
			     (vector-ref (car v-trail) 2))) ; old value
		  (else (set-cdr! (vector-ref (car v-trail) 0) ; cdr of cell
			     (vector-ref (car v-trail) 2)))) ; old value
	    (v-trail-previous)
	    (undo-v-trail))
     (v-trail-previous)))
(define (display-substitution trail)
; displays the substitutions back to the previous trail stop
 (if (caar trail)
     (begin (display-substitution (cdar trail)) 
            (newline)
            (display (cdr (caar trail))) (display '/) 
            (display (caaar trail)))))
     
(define (display-ref term)
 (if (free-variable term)
     (display (var-name term))
     (if (variable term)
         (display-ref (ref term))
         (begin (display (functor-symbol (car term))) 
                (if (pair? (cdr term))
		    (begin (display "(") 
			   (display-ref-list (cdr term))
			   (display ")")))))))

(define (display-ref-list terms)
 (if (pair? terms)
     (begin (display-ref (ref (car terms)))
            (if (pair? (cdr terms))
	        (begin (display ",") (display-ref-list (cdr terms)))))
     '#\ ))

(define (display-subgoal-list subgoal-list)
 (if (pair? subgoal-list)
     (begin (display-subgoal (car subgoal-list)) (newline)
            (display-subgoal-list (cdr subgoal-list)))))

(define (display-subgoal subgoal)
 (display (subgoal-sign subgoal)) 
 (display (predsym-symbol (subgoal-pred subgoal))) 
 (if (> (predsym-arity (subgoal-pred subgoal)) 0) 
     (begin (display "(")
            (display-ref-list (subgoal-terms subgoal)) 
	    (display ")"))))

(define (display-subgoals subgoals offset)
 (if (< offset (- (vector-length subgoals) 1))
     (begin (display-subgoal (vector-ref subgoals offset))
            (display-subgoals subgoals (+ offset 1)))))

(define (display-copies all-copies new-copies)
 (if (not (eq? all-copies new-copies))
     (begin (display-subgoals (copy-subgoals (car all-copies)) 0)
;            (display-copy-constraints (copy-subgoals (car all-copies)))
	    (newline) (display "        ")
            (display-copies (cdr all-copies) new-copies))))

(define (display-clauses clauses)
 (if (pair? clauses)
     (begin (display-clause (car clauses)) (newline)
            (display-clauses (cdr clauses)))))

(define (display-clause cl)
 (newline) (display "clause ") (display (clause-number cl)) (display ": ") 
 (display-subgoals (clause-generic cl) 0) 
 (if (not (eq? (clause-copies cl) (clause-all-copies cl)))
     (begin (newline) (display "copies: ") 
            (display-copies (clause-all-copies cl) (clause-copies cl)))))

;(define (display-copy-constraints subgoals)
; (display-constraints 
;  (term-variables (subgoals-termlists-appended subgoals 0))))

 (define (display-proof-clauses clauses)
 (if (pair? clauses)
     (begin (if (not (eq? (clause-all-copies (car clauses))
                          (clause-copies (car clauses))))
                (display-clause (car clauses)))
            (display-proof-clauses (cdr clauses)))
     (newline)))

(define (display-trail-content trail)
; displays the contents of the entire trail
 (begin (if (cdar trail)
            (display-trail-content (cdar trail)))
        (print (caar trail) "")))

(define (display-inferences depth)
 (display "Level ") (display depth)
 (display ": successful/failed inferences: ") 
 (display successful-inferences) (display "/") 
 (display failed-inferences)
 (display " (") (display constraint-fails) (display ") ")
 (if sgindex
     (begin (display nr-subgoals) (display "/") (display successful-matchings)
	    (set! nr-subgoals 0) (set! successful-matchings 0)))
 (newline)
 (set! total-successful-inferences (+ total-successful-inferences 
                                      successful-inferences))
 (set! total-failed-inferences (+ total-failed-inferences 
				  failed-inferences))
 (set! total-constraint-fails (+ total-constraint-fails constraint-fails))
 (set! successful-inferences 0) (set!  failed-inferences 0)
 (set! constraint-fails 0)
)


(define (display-links subgoal)
 (newline) (display-links-subgoal (subgoal-links subgoal)))

(define (display-links-subgoal links)
 (if (pair? links)
     (if (eq? 'reduction (car links))
         (display-links-subgoal (cdr links))
         (begin (display (clause-number (caar links))) 
                (display " ") (display (cdar links))
                (newline) (display-links-subgoal (cdr links))))))

(define (display-constraints variables)
 (if (pair? variables)
     (begin (display (var-name (car variables)))
	    (if (not (free-variable (car variables)))
		(begin (display "/") (display-ref (car variables))))
	    (display ":  ")
	    (display-constraints1 (var-constraints (car variables)))
	    (newline)
	    (display-constraints (cdr variables)))
     ""))

(define (display-constraints1 subconstraint)
  (if subconstraint
      (begin (display-ref (vector-ref subconstraint 0))
	     (display "[")
             (display (vector-ref (vector-ref subconstraint 1) 0))
	     (display ",")
             (display (vector-ref (vector-ref subconstraint 1) 1))
	     (display "]  ")
	     (display-constraints1 (vector-ref subconstraint 2)))))
 
(define (display-v-trail point)
 (display (vector-ref (car point) 1))
 (display " ")
 (display (vector-ref (car point) 2))
 (if (eq? point v-trail)
     ""
     (begin (display " | ")
	    (display-v-trail (cdr point)))))






;;;*************************************************************************
;; ----------------------------------------------
;; START CLAUSE SELECTION AND ITERATIVE DEEPENING
;; ----------------------------------------------

(define (all-negated-conjectures clauses)
 (if (pair? clauses)
     (if (or (has-positive-literal (clause-generic (car clauses))
                                   (- (clause-length (car clauses)) 1))
             (eq? (clause-type (car clauses)) 'conjecture))
         (all-negated-conjectures (cdr clauses))
         #f)
     #t))
         
(define (has-positive-literal subgoals index)
 (if (>= index 0)
     (if (eq? (subgoal-sign (vector-ref subgoals index)) '+) 
         #t
         (has-positive-literal subgoals (- index 1)))
     #f))

(define (conjecture-clauses clauses)
 (if (pair? clauses)
     (if (eq? (clause-type (car clauses)) 'conjecture)
         (cons (car clauses) (conjecture-clauses (cdr clauses)))
         (conjecture-clauses (cdr clauses)))
     '()))

(define (negated-clauses clauses)
 (if (pair? clauses)
     (if (not (has-positive-literal (clause-generic (car clauses))
                                    (- (clause-length (car clauses)) 1)))
         (cons (car clauses) (negated-clauses (cdr clauses)))
         (negated-clauses (cdr clauses)))
     '()))

(define (free-signature-hyp-con clauses)
 (if (pair? clauses)
     (begin (if (or (eq? (clause-type (car clauses)) 'conjecture)
                    (eq? (clause-type (car clauses)) 'hypothesis))
                (free-signature-weights (clause-signature (car clauses))))
            (free-signature-hyp-con (cdr clauses)))))

(define (remove-free-signature-clauses clauses)
 (if (pair? clauses)
     (begin (set-clause-signature (car clauses)
                    (remove-free-signature (clause-signature (car clauses))))
            (remove-free-signature-clauses (cdr clauses)))))

(define (remove-free-signature sigs)
 (if (pair? sigs)
     (if (> (signature-weight (car sigs)) 0)
         (cons (car sigs) (remove-free-signature (cdr sigs)))
         (remove-free-signature (cdr sigs)))
     '()))

(define (remove-lemmas clauses lemmas)
 (if (pair? clauses)
     (if (eq? (clause-type (car clauses)) 'lemma)
	 (remove-lemmas (cdr clauses) (cons (car clauses) lemmas))
	 (cons (car clauses) (remove-lemmas (cdr clauses) lemmas)))
     (begin (set-formula-lemmas formula lemmas)
	    '())))

(define (setheo args)
 (update-defaults args)
 (set! filename (car (last args)))
 (read-tptp)
 (set-formula-clauses formula (remove-lemmas (formula-clauses formula) '()))
 (in-lemma-index-lemmas (formula-lemmas formula))
 (if clause-reordering (sort-clauses (formula-clauses formula)))
 (if equality (if (= nr-eq-literals 0)
                  (set! equality #f)
                  (begin (display "Equality handling") (newline))))
 (make-connection-graph (formula-clauses formula) (formula-clauses formula))
 (if equality (make-eq-links (formula-clauses formula)))
; (set! start-clauses (formula-clauses formula))
 (if (or norel (all-negated-conjectures (formula-clauses formula)))
     (begin (newline) (display "all negated are conjectures")
     (set! start-clauses (negated-clauses (formula-clauses formula)))
     )
     (begin (newline) (display "some negated are no conjectures")
     (set! start-clauses (conjecture-clauses (formula-clauses formula)))
     )
)
 (if signature-bound
     (begin (update-pred-numbers predicate-list)
            (set! sig-vector (make-vector (+ nr-functors nr-predicates 1) #f))
            (set! sig-trail (make-vector (+ nr-functors nr-predicates)))
            (if (> nr-axioms 0)
                (begin (free-signature-hyp-con (formula-clauses formula))
                   (remove-free-signature-clauses (formula-clauses formula))))
     ))
 (newline)
 (search-proof start-clauses)
)

(define (search-proof clauses) 
 (if (pair? clauses)
     (begin ;(display-clause (car clauses))
            (set! difference-counter 0)
            (extension #f (car clauses) #f depth 0)
            (search-proof (cdr clauses)))
     (begin (display-inferences depth)
	    (if bound-fail
                (begin (set! bound-fail #f)
                       (if (= depth final-depth)
                           (begin (newline)
                                  (display "     ***** BOUND REACHED *****")
                                  (newline) ; (exit 0)
)
		           (begin (set! depth (+ depth 1))
                                  (search-proof start-clauses))))
                (begin (newline) 
		   (display "     ***** FORMULA IS SATISFIABLE *****")
		   (newline) (newline)
	   (display "Total successful/failed inferences(constraints): ") 
		   (display total-successful-inferences) (display "/")
		   (display total-failed-inferences) (display " (")
		   (display total-constraint-fails) (display ")") (newline)
;		   (exit 0)
		   #f)))))

;;;END *********************************************************************

;; INDEX 
;; -----
;; Description.
;; An index stores a set of terms in such a way that each term occurs in only
;; one variant. The checking for the existence of a variant and the insertion
;; of a new variant can be done very efficiently.
;;
;; Index data structures.
;; INDEX_VARIABLE ::= POSINT 
;; index variables are identical if their POSINTs are identical
;; INDEX_FUNC ::= #( ATOM ARITY NUMBER )
;; INDEX_HEAD ::= INDEX_VARIABLE | INDEX_FUNC
;; INDEX_END ::= ( INDEX_HEAD )
;; by simultaneous induction:
;; OR_NODE ::= ( AND_NODE . OR_NODES )
;; OR_NODES ::= () | ( OR_NODE . OR_NODES )
;; AND_NODE ::= ( INDEX_HEAD . OR_NODES )
;;;end********************************************************************
;;
;; Definition (WELL-FORMED_OR_NODE).
;; The OR_TREE of an OR_NODE is the labelled tree representation of the
;; list structure of the OR_NODE where INDEX_HEAD and INDEX_END
;; expressions are taken as atomic. This means that the OR_TREE of an 
;; OR_NODE has the following structure: all inner nodes are unlabelled,
;; the leftmost successor node of every inner node is a leaf node labelled 
;; with an INDEX_HEAD expression, and all other leaf nodes are labelled 
;; with INDEX_END expressions.
;; An OR_NODE is well-formed if every INDEX_END i has the following 
;; property: the number of predecessor nodes of i equals the sum of the 
;; ARITIES in the INDEX_FUNC expressions which are at the leftmost brother 
;; nodes of i and its predecessor nodes.

;; Clearly, every INDEX_END of a WELL-FORMED_OR_NODE represents a 
;; well-formed logical term.
;; 
;; INDEX ::= () | ( WELL-FORMED_OR_NODE . INDEX )
;; INDEX_ROOT ::= ( INDEX . NATNUM )
;; NATNUM is the maximal number of distinct variables in an index term 
;;
;; Note that a propositional index is simply #t or #f, depending on whether
;; a propositional literal occurs in the index or not

;; ---------------------------------------------------------
;; PROCEDURES FOR CHECKING WHETHER A TERMLIST IS IN AN INDEX
;; ---------------------------------------------------------

(define (is-in-index-root? termlist term index-root-in)
; checks whether a nonempty termlist with first car term is in index;
; the procedure uses the standard binding and trailing mechanisms for
; associating term variables with index variables
 (if (pair? (car index-root-in)) ; index is nonempty
     (begin (set! index (car index-root-in)) (set! nth-variable 0)
	    (set! max-variables (cdr index-root-in)) (trail-stop) 
	    (if (is-in-or-node? termlist term)
		(begin (unbind-vars) #t)
		(begin (unbind-vars) #f)))
     #f))

(define (is-in-or-node? termlist term)
 (if (or (integer? term)         ; term is a variable
	 (and (free-variable term) 
              (begin (set! nth-variable (+ 1 nth-variable))
		     (and (<= nth-variable max-variables)
                          (begin (bind term nth-variable)
                                 (set! term nth-variable))))))
     (and (number? (caar index)) ; first entry in index is a variable
	  (is-in-nodes-var? term (caar index)) ; note: term is variable nr 
          (or (eq? '() (cdr termlist)) 
	      (is-in-or-node? (cdr termlist) (ref-int (cadr termlist)))))
     (and (is-in-nodes-func? (car term))
          (or (eq? '() (cdr term))
              (is-in-or-node? (cdr term) (ref-int (cadr term))))
          (or (eq? '() (cdr termlist))
              (is-in-or-node? (cdr termlist) (ref-int (cadr termlist)))))))

(define (is-in-nodes-var? var-number next-var-number)
 (if (= var-number next-var-number) 
     (set! index (cdar index))
     (and (< var-number next-var-number)
	  (pair? (cdr index)) 
          (number? (caadr index))
	  (begin (set! index (cdr index)) 
                 (is-in-nodes-var? var-number (caar index))))))

(define (is-in-nodes-func? funcsymb)
 (if (number? (caar index)) ; top of first OR_NODE is a variable
     (and (pair? (cdr index))
          (begin (set! index (cdr index)) (is-in-nodes-func? funcsymb)))
     (is-in-nodes-func1? (functor-number funcsymb) ; no variable
                         (functor-number (caar index)))))

(define (is-in-nodes-func1? number next-number) 
; auxiliary to is-in-nodes-func1?
 (if (= number next-number) 
     (set! index (cdar index)) ; funcsymb in index
     (and (< number next-number)
          (pair? (cdr index)) 
	  (begin (set! index (cdr index))
                 (is-in-nodes-func1? number (functor-number (caar index)))))))

;; --------------------------------------------------------------------
;; PROCEDURES FOR CHECKING WHETHER AN INDEX TERMLIST MATCHES A TERMLIST
;; --------------------------------------------------------------------

(define (match-index-root? termlist term index-root-in)
; checks whether a nonempty termlist with car term has a match from
; a termlist in the index;
; the procedures use the trailing mechanisms for keeping track of the
; parts of the input termlist that have been delayed
 (set! in-index? #f)
 (if (pair? (car index-root-in)) ; index is nonempty
     (begin (trail-stop)
	    (match-or-node? (car index-root-in) termlist term)
	    (trail-previous)))
 in-index?)

(define (match-or-node? or-node termlist term)
;(newline) (display "V: ") (display-ref-list termlist)
;(display " vs ") (display (caar or-node))
 (let (tl undo-binding)
   (if (integer? (caar or-node)) ; first entry in index is a variable
       (if (or (and (not (vector-ref index-bindings (caar or-node))) ; free
		    (begin (vector-set! index-bindings (caar or-node) term)
			   (set! undo-binding (caar or-node)) #t)) ; bound    
	       (ref-equal (vector-ref index-bindings (caar or-node)) term))
	   (begin (cond ((pair? (cdr termlist))
			 (match-or-node? (cdar or-node) (cdr termlist) 
					 (ref (cadr termlist))))
			((trail-content) ; remaining termlist(s) to be checked
			 (set! tl (trail-content)) (trail-previous)
			 (match-or-node? (cdar or-node) tl (ref (car tl)))
			 (push-trail tl)
;			 (newline) (display "VRT: ") 
;			 (display-ref-list (trail-content))
			 ); rearrange trail-pointer
			(else (set! in-index? #t))) ; matching term found
		  (if (integer? undo-binding) 
		      (vector-set! index-bindings undo-binding #f))
		  (if (and (not in-index?) (pair? (cdr or-node))) ; deep back-
		      (match-or-node? (cdr or-node) termlist term))) ; tracking
	   (if (pair? (cdr or-node)) ; shallow backtracking
	       (match-or-node? (cdr or-node) termlist term)))
       (if (non-variable term) ; term must be no variable, too
	   (match-or-node-func? or-node termlist term
				(functor-number (car term))
				(functor-number (caar or-node)))))))

(define (match-or-node-func? or-node termlist term funcnr-term funcnr-next)
;(newline) (display "F: ") (display-ref-list termlist)
;(display " vs ") (display (caar or-node))
(cond ((< funcnr-term funcnr-next)) ; no matching term exists in or-node
       ((> funcnr-term funcnr-next) ; try next function symbol in index
        (if (pair? (cdr or-node))
	    (match-or-node-func? (cdr or-node) termlist term funcnr-term
				 (functor-number (caadr or-node)))))
       ; below: top function symbols are equal 
       ((pair? (cdr term)) ; term is complex
	(if (pair? (cdr termlist))
	    (begin (push-trail (cdr termlist))
;	       (newline) (display "IT: ") (display-ref-list (trail-content))
	       (match-or-node? (cdar or-node) (cdr term) (ref (cadr term)))
	       (trail-previous)) ; rearrange trail-pointer
	    (match-or-node? (cdar or-node) (cdr term) (ref (cadr term)))))
       ((pair? (cdr termlist)) 
	(match-or-node? (cdar or-node) (cdr termlist) (ref (cadr termlist))))
       ((trail-content) ; remaining termlist(s) to be checked
	(set! termlist (trail-content)) (trail-previous)
	(match-or-node? (cdar or-node) termlist (ref (car termlist)))
	(push-trail termlist) ; rearrange trail-pointer
;	(newline) (display "FRT: ") (display-ref-list (trail-content))
       )
       (else (set! in-index? #t))))


;; -------------------------------------------------
;; PROCEDURES FOR INSERTING A TERMLIST INTO AN INDEX
;; -------------------------------------------------

(define (in-index-root termlist term index-root-in) 
; top procedure for inserting a nonempty TERMLIST with car TERM into an
; INDEX-ROOT
 (set! index-root index-root-in) (set! max-variables (cdr index-root))
 (set! nth-variable 0) (set! var-assoc-nr? #f)
 (if (pair? (car index-root)) ; index is nonempty
     (begin (set! in-index? #t) (set! index (car index-root))
            (in-or-node termlist term))
     (begin (set! in-index? #f) ; index-root is empty
        (if (variable term) ; first term is a variable
            (begin (set-car! index-root (list (list (var-assoc-nr term))))
		   (set! index (caar index-root)))
            (begin (set-car! index-root (list (list (car term))))
		   (set! index (caar index-root))
		   (if (pair? (cdr term)) 
		       (in-and-node (cdr term) (ref (cadr term))))))
	(if (pair? (cdr termlist)) 
	    (in-and-node (cdr termlist) (ref (cadr termlist))))))
 (set-cdr! index-root max-variables)
 (if (> max-variables (vector-length index-bindings))
     (set! index-bindings (make-vector (+ 1 max-variables) #f)))
 in-index?)

(define (in-or-node termlist term)
; main procedure for inserting a nonempty TERMLIST with car TERM into an
; INDEX-ROOT
; global variables:
; nth-variable: NATNUM, the number of the last scanned variable;
; max-variables: NATNUM; if a term with more distinct variables as in the
; index is processed, max-variables is set to the new value;
; var-assoc-nr?: #f or list of already seen input variables in termlist; 
; whenever a new variable is encountered in termlist, the list ( variable )
; is destructively attached at the end of var-assoc-nr?;
; var-assoc-nr? must be set to #f before the inital call to in-index; 
; in-index?: #t or #f depending on whether the TERM is in INDEX or not; 
; must be set to #t before the inital call to in-index
; index: is always set to the current OR_NODE or the last inserted AND_NODE;
; SIDE EFFECTS: 
; 1. in-index?, nth-variable, max-variables, var-assoc-nr?, index;
; 2. index-root is possibly modified, by integrating TERMLIST 
 (if (variable term) ; first term is a variable: finds or inserts symbol at
     (begin (var-assoc-nr term) ; the right position in the current index
            (if (number? (caar index)) ; first entry in index is a variable
                (in-nodes-var nth-variable (caar index))
                (begin (set! in-index? #f) ; since variables come first
                       (set-cdr! index (cons (car index) (cdr index))) 
                       (set-car! index (list nth-variable))
                       (set! index (car index)))))
     (begin ; term is no variable: finds or inserts the top function symbol 
            (in-nodes-func (car term))
            (if (pair? (cdr term))
                (if in-index? 
		    (in-or-node (cdr term) (ref (cadr term)));index is: OR_NODE
		    (in-and-node (cdr term) (ref (cadr term))))))) ; : AND_NODE
 (if (pair? (cdr termlist))
     (if in-index? 
         (in-or-node (cdr termlist) (ref (cadr termlist))) ; index is: OR_NODE 
         (in-and-node (cdr termlist) (ref (cadr termlist)))))) ;   : AND_NODE

(define (in-nodes-var var-number next-var-number)
;; checks whether the var-number-th universal variable in a respective
;; term position is in index; 
;; if not, a new OR_NODE is inserted at the right index position
 (cond ((= var-number next-var-number) (set! index (cdar index)))
       ((< var-number next-var-number) (set! in-index? #f)
        (set-cdr! index (cons (car index) (cdr index))) 
        (set-car! index (list var-number))
        (set! index (car index)))
       ((and (pair? (cdr index)) 
                    (number? (caadr index))) ; next OR_NODE is a variable
        (set! index (cdr index)) (in-nodes-var var-number (caar index)))
       (else (set! in-index? #f) 
             (set-cdr! index (cons (list var-number) (cdr index)))
             (set! index (cadr index)))))

(define (in-nodes-func funcsymb)
; checks whether a function symbol in a respective position is in index;
; at the beginning, the list of variables in the index is skipped;
; depending on whether the function symbol is in the index, a new
; OR_NODE adn AND_NODE is inserted into the index at the right position
; (ie function symbols are inserted in increasing order of their numbers);
; when entering the procedure, the global variable index is an OR_NODE;
; at the end of the computation, index is set to 
; * the next actual OR_NODE       if the function symbol is in the index
; * the last inserted AND_NODE    otherwise
; (newline) (display "INDEX: ") (display index)
; (newline) (display "FUNCSYM: ") (display funcsymb)
 (if (number? (caar index)) ; top of first OR_NODE is a variable
     (if (pair? (cdr index))
         (begin (set! index (cdr index)) (in-nodes-func funcsymb))
         (begin (set! in-index? #f) (set-cdr! index (list (list funcsymb)))
                (set! index (cadr index))))
     (in-nodes-func1 funcsymb (functor-number funcsymb) ; no variable 
		     (functor-number (caar index)))))
     
(define (in-nodes-func1 funcsymb number next-number) 
; auxiliary to in-nodes-func
 (cond ((= number next-number) (set! index (cdar index))) ; funcsymb in index
       ((< number next-number) (set! in-index? #f)
        (set-cdr! index (cons (car index) (cdr index))) 
        (set-car! index (list funcsymb))
        (set! index (car index))) ; below: (> number next-number)
       ((pair? (cdr index)) (set! index (cdr index)) ; next OR_NODE 
        (in-nodes-func1 funcsymb number (functor-number (caar index))))
       (else (set! in-index? #f) ; last OR_NODE reached
        (set-cdr! index (list (list funcsymb)))
        (set! index (cadr index)))))

(define (in-and-node termlist term)
; (if (= nr-sub;goals 17)-in-nodes-var? var-number 
;     (begin (newline) (display "IN-AND-NODE ENTERED: ")
;	    (display-ref-list termlist)
;	    (display index)))
; inserts nonempty termlist with first term into the current and-node
; when entering the procedure, index must be set to an AND_NODE
 (if (variable term) ; first term is a variable
     (begin (set-cdr! index (list (list (var-assoc-nr term))))
	    (set! index (cadr index))) ; index set to next AND_NODE
     (begin (set-cdr! index (list (list (car term))))
	    (set! index (cadr index)) ; index set to next AND_NODE
            (if (pair? (cdr term)) 
		(in-and-node (cdr term) (ref (cadr term))))))
 (if (pair? (cdr termlist))
     (in-and-node (cdr termlist) (ref (cadr termlist)))))

(define (var-assoc-nr variable)
;; performs association of term variables with index variables
(if var-assoc-nr?
    (begin (set! nth-variable (nth-occ-insert variable var-assoc-nr? 1))
           (if (> nth-variable max-variables) 
	       (set! max-variables nth-variable))
           nth-variable)
    (begin (set! var-assoc-nr? (list variable))
           (if (= max-variables 0) (set! max-variables 1))
           (set! nth-variable 1) 1)))

(define (nth-occ-insert expr list1 position) 
;; INPUT: list1: list of length m > 0
;;        position: NATNUM, initially 1
;; OUTPUT: if expr eq-occurs in list1, the first such position k is given,
;;         else m+1 is outputted and
;;         SIDE EFFECT: expr is destructively attached to the end of list1
 (cond ((eq? expr (car list1)) position)
       ((pair? (cdr list1)) (nth-occ-insert expr (cdr list1) (+ 1 position))) 
       (else (set-cdr! list1 (list expr)) (+ 1 position))))

(define (print-and-node prefix and-node) 
 (if (vector? (car and-node))
     (set! prefix (cons (functor-symbol (car and-node)) prefix))
     (set! prefix 
           (cons (string-append "x" (number->string (car and-node))) prefix)))
 (if (pair? (cdr and-node))
     (print-or-node prefix (cdr and-node))
     (begin (newline) (display (print-list (reverse prefix))))))

(define (print-or-node prefix or-node)
 (if (pair? or-node)
     (begin (print-and-node prefix (car or-node))
            (print-or-node prefix (cdr or-node)))))

(define (print-list list1)
 (if (pair? list1)
     (begin (display (car list1)) (display " ") (print-list (cdr list1)))
     '#\ ))

; Examples for testing

(define index-root-in (cons '() 0))
 
;(set! filename "test1.p") (read-tptp)
;(begin (sort-clauses (formula-clauses formula)) 'done)
;(make-connection-graph (formula-clauses formula) (formula-clauses formula))
;(define generic1 (clause-generic (car (formula-clauses formula))))
;(define s0 (vector-ref generic1 0)) (define t0 (subgoal-terms s0))
;(define s1 (vector-ref generic1 1)) (define t1 (subgoal-terms s1))
;(define s2 (vector-ref generic1 2)) (define t2 (subgoal-terms s2))
;(define s3 (vector-ref generic1 3)) (define t3 (subgoal-terms s3))
;(define s4 (vector-ref generic1 4)) (define t4 (subgoal-terms s4))
;(define s5 (vector-ref generic1 5)) (define t5 (subgoal-terms s5))
;(define s6 (vector-ref generic1 6)) (define t6 (subgoal-terms s6))
;(define s7 (vector-ref generic1 7)) (define t7 (subgoal-terms s7))
;(define s8 (vector-ref generic1 8)) (define t8 (subgoal-terms s8))

;(define (inindex termlist)
; (in-index-root termlist (ref (car termlist)) index-root-in))

;(define (inindex? termlist)
; (is-in-index-root? termlist (ref (car termlist)) index-root-in))

;(define (matchindex? termlist)
; (match-index-root? termlist (ref (car termlist)) index-root-in))

;(define (printindex) (print-or-node '() (car index-root-in)))

(define (in-lemma-index-lemmas lemmas)
 (if (pair? lemmas)
     (begin 
       (if (= 1 (clause-length (car lemmas)))
	   (let ((subgoal (vector-ref (clause-generic (car lemmas)) 0)))
	     (if (> (predsym-arity (subgoal-pred subgoal)) 0)
		 (if (eq? '+ (subgoal-sign subgoal))
		     (in-index-root (subgoal-terms subgoal)
			      (car (subgoal-terms subgoal))
			      (predsym-pos-index (subgoal-pred subgoal)))
		     (in-index-root (subgoal-terms subgoal)
			      (car (subgoal-terms subgoal))
			      (predsym-neg-index (subgoal-pred subgoal))))
		 (if (eq? '+ (subgoal-sign subgoal))
		     (set-predsym-pos-index (subgoal-pred subgoal) #t)
		     (set-predsym-neg-index (subgoal-pred subgoal) #t)))))
       (in-lemma-index-lemmas (cdr lemmas)))))

(define (print-predicate-lemma-indeces predicates)
 (if (pair? predicates)
     (begin (newline) (newline) 
	    (display "+ ") (display (predsym-symbol (car predicates)))
	    (display ": ---------------------------")
	    (print-or-node '() (car (predsym-pos-index (car predicates))))
	    (newline) (newline) 
	    (display "- ") (display (predsym-symbol (car predicates)))
	    (display ": ---------------------------")
	    (print-or-node '() (car (predsym-neg-index (car predicates))))
	    (print-predicate-lemma-indeces (cdr predicates)))))

(define (find-matching-lemma subgoal)
 (if (> (predsym-arity (subgoal-pred subgoal)) 0)
     (if (eq? '- (subgoal-sign subgoal))
	 (match-index-root? (subgoal-terms subgoal)
			    (ref (car (subgoal-terms subgoal)))
			    (predsym-pos-index (subgoal-pred subgoal)))
	 (match-index-root? (subgoal-terms subgoal)
			    (ref (car (subgoal-terms subgoal)))
			    (predsym-neg-index (subgoal-pred subgoal))))
     (if (eq? '- (subgoal-sign subgoal))
	 (predsym-pos-index (subgoal-pred subgoal))
	 (predsym-neg-index (subgoal-pred subgoal)))))

