#ifndef SEM_CONSTR_H
#define SEM_CONSTR_H

#include "machine.h"



/********** procedures for the creation of the sem_constr-tree ************/

sem_junct*  create_sem_junct();
/* sem_junct*  create_sem_junct(int junctType);

   create new sem_junct on the constraint-stack and initialize it
   It is by default a root junctor.
*/

void add_sem_junct();
/* void add_sem_junct(sem_junct* parentJunct, sem_junct* myJunct);

   Add myJunct to the sub_constr_list of parentJunct. */

sem_junct*  add_new_sem_junct();
/* sem_junct*  add_new_sem_junct(sem_junct* parentJunct, int junctType);
 *
 * create new sem_junct on the constraint-stack,
 * initialize it and add it to the sub_constr_list of parentJunct.
 * If parentJunct == NULL a new root node is being created 
 */

void  add_to_sub_constr_list();
/* void  add_to_sub_constr_list(sem_junct* outParentJunct, 
                                sem_child* child, 
				int type,
				int position);
 *
 * allocate new sem_constr_list element on the constraint stack,
 * initialize it (with child and type),
 * and insert it as first element into the sub_constr_list of
 * outParentJunct */ 

void  insert_in_elem_constr_lists(sem_junct*  junct);
/* void  insert_in_elem_constr_lists(sem_junct*  junct); */


/******* procedures for the attribute-handling in the sem_constr-tree ********/

void  remove_junct_with_arity_one();
/* void  remove_junct_with_arity_one(sem_junct*  junct); */

int is_sem_elem_constr_alive();
/* int is_sem_elem_constr_alive(sem_elem_constr* elemConstr);
 *
 * return 1, if the constraint is neither deleted nor deactivated 
 * else      mark the whole branch of this constraint and return 0
 */ 

int is_sem_junct_alive();
/* int is_sem_junct_alive(sem_junct*  myJunct);
 * is_sem_junct_alive
 * return TRUE, if the constraint is neither deleted nor deactivated 
 * else      mark the whole branch of this constraint and return FALSE
 */

void deactivate_sem_elem_constr();
/* void deactivate_sem_elem_constr(sem_elem_constr* elemConstr);
 *
 * Deactivate living elemConstr, i.e. it can't be violated any more
 */

int delete_sem_elem_constr();
/* int delete_sem_elem_constr(sem_elem_constr* elemConstr);
 *
 * Delete living elemConstr, i.e. it can't be fulfilled any more
 */

void generate_sem_elem_constraints_from_pu_trail();
/* void generate_sem_elem_constraints_from_pu_trail(sem_junct*  parentJunct, 
                                                    int         type,  
                                                    int         insertFlag);      
 *
 * generate new elementary subconstraints from bindings, which 
 * have been written on the PU-trail during the protocol unification.
 */

void decompose_sem_elem_constr();
/* void decompose_sem_elem_constr(sem_elem_constr* elemConstr);
 *
 * decompose elementary constraints which still can be fulfilled or violated
 * i.e. generate new elementary subconstraints from bindings, which 
 * have been written on the PU-trail during the protocol unification.
 * The original elementary constraint will be deactivated 
 */



/************** deterministic instantiation *************************/

int try_det_inst();
/* int try_det_inst();
 */


#endif /* SEM_CONSTR_H */
