#ifndef SEM_TOOLS_H
#define SEM_TOOLS_H



/*********** procedures for list handling **************/

int is_list();
/* int is_list(WORD* list); */

void check_list();
/* void check_list(WORD* list);
  
  Is list a list? If not, signal an error */

void check_nonempty_list();
/* void check_nonempty_list(WORD* list);

   Is list a non-empty list? If not, signal an error */

int is_list_empty();
/* int is_list_empty(WORD* list); */

int list_length();
/* int list_length(WORD* list); */

sem_constr_list  push();
/* sem_constr_list  push(sem_junct*  junct, sem_constr_list  list); */

sem_constr_list  pop();
/* sem_constr_list  pop(sem_constr_list  list); */

int  member();
/* int  member(sem_junct*  junct, sem_constr_list  list); */


/*********** procedures for constraint input **************/

void extract_constraint_sides();
/* void extract_constraint_sides(WORD*   constraint, 
                                 WORD**  leftListPtr, 
				 WORD**  rightListPtr);
   
   constraint must be a list containing (exactly) two non-empty lists,
   if not: error! Extract these an put 'em in leftListPtr and rightListPtr */



/*********** procedures for constraint and variable output **************/

void print_sem_elem_constr();
/* void print_sem_elem_constr(sem_elem_constr* self, int indent);
 * 
 * Display elem constr self with the indentation indent */

void print_sem_constr_tree();
/* void print_sem_constr_tree(sem_junct* self, int indent);
 * 
 * Display junctor self and its sons with the indentation indent */

void print_sem_constr_trees_at_variable();
/* void print_sem_constr_trees_at_variable(WORD* var);
 *
 * Display all sem_constr_trees which contain sem_elem_constrs of var
 */ 

void  print_syn_constrs_of_variable();
/* void  print_syn_constrs_of_variable(WORD*  var); */

void  print_instantiations(); 
/* void  print_instantiations(WORD*  varBlock, 
                              int    nrVars, 
			      int    GETPTR_flag); */



/*********** helper procedures  **************/

sem_constr_list add_to_junct_list();
/* sem_constr_list add_to_junct_list(sem_constr_list junctList, 
                                     sem_junct* junct); */

void free_sem_constr_list();
/* void free_sem_constr_list(sem_constr_list list); */

int junct_list_length();
/* int junct_list_length(sem_constr_list junctList); */

int junct_list_contains();
/* int junct_list_contains(sem_constr_list junctList, 
                           sem_junct* junct); */

sem_junct* get_root_junct();
/* sem_junct* get_root_junct(sem_elem_constr* constr); */
/* Find the root junct of a given (elementary) constraint
   Helper function */

sem_constr_list compute_list_of_root_nodes_at_variable();
/* sem_constr_list  compute_list_of_root_nodes_at_variable(WORD* var); */

WORD  *copy_instantiations();
/* WORD  *copy_instantiations(choice_pt *choicePoint); */

WORD  *copy_complex_term();
/* WORD  *copy_instantiations(choice_pt *choicePoint); */

int  is_variable_relevant();
/* int  is_variable_relevant(WORD* var); */


/***** procedures for constraint evaluation *****/

int nr_of_top_level_constraints_satisfied_by_elem_constr();
/* int nr_of_top_level_constraints_satisfied_by_elem_constr(sem_elem_constr* constr); */

int  dag_tc_of_sem_elem_constr();
/* int  dag_tc_of_sem_elem_constr(sem_elem_constr* constr, 
			          int              limit); */


/***** procedures for variable evaluation *****/

int  nr_of_constraints_at_variable();
/* int  nr_of_constraints_at_variable(WORD* myVar); */

int  max_nr_of_appearances_in_same_parent_AND_node(WORD* var);
/* int  max_nr_of_appearances_in_same_parent_AND_node(WORD* var); */



/***** procedures for input administration *****/

void  adapt_local_consistency_check();
/* void  adapt_local_consistency_check(); */

int  check_local_consistency();
/* int  check_local_consistency(choice_pt*  chpt,
                                int         nrTreesToCheck); */

void  attach_to_imagined_root_node();
/* void  attach_to_imagined_root_node(sem_junct* junct); */



/***** procedures for lcanl/gcanl generation *****/

a_list*  generate_canl();
/* a_list*  generate_canl(history_list  historyForCanl,
		          history_list  history,
		          a_list*       lcanl_list); */

a_list*  init_canl();
/* a_list*  init_canl(int      length_bt_list,
	              a_list*  canl_list); */

a_list*  insert_in_canl_list();
/* a_list*  insert_in_canl_list(a_list*  canl_list); */

#endif
