/******************************************************/
/*    S E T H E O                                     */
/*                                                    */
/* FILE: i_disconstr.c                                */
/* VERSION:                                           */
/* DATE:                                              */
/* AUTHOR:                                            */
/* NAME OF FILE:                                      */
/* DESCR:                                             */
/* MOD:                                               */
/* BUGS:                                              */
/******************************************************/

#include <stdio.h>
#include "tags.h"
#include "machine.h"
#include "deref.h"
#include "errors.h"
#include "disp.h"
#include "unitlemma.h"
#include "sem_consistency.h"
#include "sem_constr.h"
#include "sem_tools.h"
#include "symbtab.h"
#include "univ_constr.h"

#define INSTR_LENGTH		2
#define LOCAL_DEBUG2



/*************** creation of the sem_constr-tree ******************/

instr_result i_disconstr()
{
   WORD*            topList;
   WORD*            constraint;
   WORD*            leftList;
   WORD*            rightList;
   sem_junct*       rootJunct;
   int              disconstrSuccess = FALSE;

#ifdef LOCAL_DEBUG
   printf("Processing constraints of clause %d ... ", 
	  GETNUMBER(bp->lcbptr->clause_nr));
   fflush(stdout);
#endif

   topList = deref(ARGV(0), bp);
   
   check_list(topList);
   
   /* create rootJunct on the constraint-stack
    * rootJunct is the disjunctor between the equality-subconstraints */ 
   rootJunct = add_new_sem_junct(NULL, JUNCT_DISJ); /* parent_junct is NULL */

   /******************************************************************/
   /* Process elements of topList:                                   */
   /******************************************************************/
   
   while (GETSYMBOL(*topList) == LISTFUNCT) {
      
      constraint = deref(topList+1, bp);    
      
      /* constraint has to be a list with two elements,
       * that are nonempty lists themselves 
       * Extract these two lists and write them into
       * leftList and rightList */
      extract_constraint_sides(constraint, &leftList, &rightList);

      if (list_length(leftList) != list_length(rightList))
	 sam_error("i_disconstr: The constraint-sides are unbalanced",
		   NULL, 2);
      
      pu_tr = pu_trail;

#ifdef LOCAL_DEBUG1
      printf("processing list equation ...\n");
      printf("leftList: ");
      disp_(stdout, leftList, bp);
      printf("\nrightList: ");
      disp_(stdout, rightList, bp);
      printf("\n");
      fflush(stdout);
#endif
      
      if (un_protoco(leftList, bp, rightList, bp, 1)) {
	 if (pu_tr > pu_trail) {
	    generate_sem_elem_constraints_from_pu_trail(rootJunct,
							ELEM_EQU,
							TRUE);
	 }
	 disconstrSuccess = TRUE;
      }
      else
	 pu_untrail();

#ifdef LOCAL_DEBUG1
      printf("\nparent of list equation:");
      print_sem_constr_tree(rootJunct, 0);
      printf("\n\n");
      fflush(stdout);
#endif
      
      topList = deref(topList+2, bp); /* next element of topList */
      check_list(topList);
   }
   
   if (!disconstrSuccess) {
      chp->consistency_fail_occurred = TRUE;
#ifdef LOCAL_DEBUG
      printf("not successful\n");
      fflush(stdout);
#endif
      return failure;
   }
   
   if (rootJunct->arity == 0) {
      /* remove rootJunct, as no constraints were generated */
      c_sp = (WORD*) rootJunct;
   }
   else {
      attach_to_imagined_root_node(rootJunct);

      /* try deterministic instantiation: */
      if (   (     rootJunct->type == JUNCT_CONJ
	      ||  (rootJunct->type == JUNCT_DISJ  &&  rootJunct->arity == 1))
	  &&  try_det_inst() == 0) {
	 /* chp->consistency_fail_occurred = TRUE; 
	    need not be set here because is always set when
	    deterministic instantiation is called */
#ifdef LOCAL_DEBUG
         printf("not successful\n");
         fflush(stdout);
#endif
	 return failure;
      }

      /* If the root node is still alive, indicate that a new
	 constraint tree has been added (useful for selective local
	 consistency check in call instruction): */
      if (is_sem_junct_alive(rootJunct)) {
	 sem_constr_added = TRUE;  /* flag, used for control of local
				      consistency check */
	 constr_tree_count++;      /* counter, used for control of local
				      consistency check */
      }
   }
   
#ifdef LOCAL_DEBUG
   printf("successful\n");
   fflush(stdout);
#endif
#ifdef LOCAL_DEBUG1
   printf("\n\nnewsemantic_root:");
   print_sem_constr_tree(semantic_root, 0);
   printf("\n\n");
   fflush(stdout);
#endif
   pc +=INSTR_LENGTH;
   return success;
}
