/******************************************************/
/*    S E T H E O                                     */
/*                                                    */
/* FILE: sem_constr.c                                 */
/* VERSION:                                           */
/* DATE:                                              */
/* AUTHOR: Rauschmayer & Sacher                       */
/* NAME OF FILE:                                      */
/* DESCR:                                             */
/* MOD:                                               */
/* BUGS:                                              */
/******************************************************/

#include <stdio.h>
#include "tags.h"
#include "machine.h"
#include "deref.h"
#include "errors.h"
#include "disp.h"
#include "unitlemma.h"
#include "sem_constr.h"
#include "sem_tools.h"
#include "univ_constr.h"
#include "unification.h"



/*************** constants ******************/

#define LOCAL_DEBUG




/******** procedures for the creation of the sem_constr-tree ************/
/* sem-constr trees have to be built top-down, i.e. start with the root node
   and finish with the leaf nodes. Use add_new_sem_junct to create junctors
   (e.g. the root junctor) and use add_new_sem_elem_constr to create
   elementary constraints (i.e. the leaf nodes) */


void add_to_sub_constr_list(outParentJunct, child, type, position)
/* allocate new sem_constr_list element on the constraint stack,
 * initialize it (with child and type),
 * and insert it as position-th element into the sub_constr_list of
 * outParentJunct */ 
sem_junct*      outParentJunct; /* parent junctor */
sem_child*      child;
int             type;           /* type of the child: junctor or
				   elemConstr */
int             position;       /* insert position */
{
  sem_constr_list  new_ele;
  sem_constr_list  predecessor;
  int              i;
  
  /* allocate new_ele on the c_stack and check for c_stack-overflow: */
  new_ele = (sem_constr_list) c_sp;
  c_sp += SIZE_IN_WORDS(sem_constr_list_ele);
  if (c_sp >= c_stack_end )
    sam_error("Constraint-Stack-Overflow", NULL, 2);

  /* initialize new_ele with child and type: */
  new_ele->child_type = type;
  new_ele->child      = child;

  /* insert new_ele as position-th element into the sub_constr_list of
     outParentJunct: */
  if (position >= 2) {
     predecessor = outParentJunct->sub_constr_list;
     for ( i=3 ; i<position ; i++ )
	predecessor = predecessor->next;
     new_ele->next = predecessor->next;
     trail_ptr((void**) &(predecessor->next));
     predecessor->next = new_ele;
  }
  else {
     new_ele->next = outParentJunct->sub_constr_list;
     trail_ptr((void**) &(outParentJunct->sub_constr_list));
     outParentJunct->sub_constr_list = new_ele;
  }

  /* increase outParentJunct's arity: */
  trail_int(&(outParentJunct->arity));
  outParentJunct->arity++;
}



/* create new sem_junct on the constraint-stack,
 * initialize it and add it to the sub_constr_list of parentJunct.
 * If parentJunct == NULL a new root node is being created */
sem_junct*  add_new_sem_junct(parentJunct, junctType)
sem_junct*     parentJunct;
int            junctType;
{
  sem_junct* outJunct;
  
  outJunct = create_sem_junct(junctType); /* create it */
  add_sem_junct(parentJunct, outJunct);   /* and add it to its parent */

  return outJunct;
}



void add_new_sem_elem_constr(type, rightSide, parentJunct, outVar, insertFlag)
/* create new sem_elem_constr on the constraint-stack and initialize it
   insert it as first element to outVar->sem_equ_constr and add it to the
   sub_constr_list of parentJunct. */
int               type;
WORD*             rightSide;
sem_junct*        parentJunct;
WORD*             outVar;
int               insertFlag;
{
  sem_elem_constr* newElemConstr;

  newElemConstr = (sem_elem_constr *) c_sp;
  c_sp += SIZE_IN_WORDS(sem_elem_constr);
  
  /* check for stack-overflow */
  if (c_sp >= c_stack_end )
    sam_error("Constraint-Stack-Overflow", NULL, 2);

  newElemConstr->type            = type;
  newElemConstr->right_side      = rightSide;
  newElemConstr->parent_junct    = parentJunct;
  newElemConstr->var             = outVar;
  newElemConstr->next            = NULL;

  if (insertFlag) {
     /* insert the new elementary constraint and trail the beginning
	of the list: */
     switch (type) {
	
     case ELEM_EQU:
	   newElemConstr->next = ((constr_block*)
				  GETPTR(*outVar))->sem_equ_constr; 
	   trail_ptr((void**) &(((constr_block*)
				 GETPTR(*outVar))->sem_equ_constr)); 
	   ((constr_block*) 
	    GETPTR(*outVar))->sem_equ_constr = newElemConstr;
	   break;
	   
     case ELEM_DIFF:
	   newElemConstr->next = ((constr_block*)
				  GETPTR(*outVar))->sem_diff_constr; 
	   trail_ptr((void**) &(((constr_block*)
				 GETPTR(*outVar))->sem_diff_constr)); 
	   ((constr_block*) 
	    GETPTR(*outVar))->sem_diff_constr = newElemConstr;
	   break;
	   
     default:
	   sam_error("add_new_sem_elem_constr: Wrong constraint type!\n",
		     NULL, 2);
     }
  }

  if (parentJunct) {
     add_to_sub_constr_list(parentJunct, 
			    (sem_child*) newElemConstr,
			    CHILD_ELEM,
			    1);
  }
  else {
     add_to_sub_constr_list(semantic_root, 
			    (sem_child*) newElemConstr,
			    CHILD_ELEM,
			    1);
  }
}



void  insert_elem_constr_in_elem_constr_list(elemConstr)
sem_elem_constr*  elemConstr;
{

   switch (elemConstr->type) {

      case ELEM_EQU:
	    elemConstr->next = ((constr_block*)
				GETPTR(*(elemConstr->var)))->sem_equ_constr; 
	    trail_ptr((void**) &(((constr_block*)
				  GETPTR(*(elemConstr->var)))->sem_equ_constr));
	    ((constr_block*)
	     GETPTR(*(elemConstr->var)))->sem_equ_constr = elemConstr;
	    break;
	   
     case ELEM_DIFF:
	    elemConstr->next = ((constr_block*)
				GETPTR(*(elemConstr->var)))->sem_diff_constr; 
	    trail_ptr((void**) &(((constr_block*)
				  GETPTR(*(elemConstr->var)))->sem_diff_constr));
	    ((constr_block*)
	     GETPTR(*(elemConstr->var)))->sem_diff_constr = elemConstr;
	    break;
	   
     default:
	   printf("elemConstr->type: %d\n", elemConstr->type);
	   sam_error("insert_elem_constr_in_elem_constr_list: Wrong constraint type!\n",
		     NULL, 2);
   }
}



void  insert_in_elem_constr_lists(junct)
sem_junct*  junct;
{
   sem_constr_list  subconstr = junct->sub_constr_list;

   while (subconstr) {
      switch (subconstr->child_type) {
      case CHILD_ELEM:
	    insert_elem_constr_in_elem_constr_list((sem_elem_constr*)
						   subconstr->child);
	    break;
      case CHILD_JUNCT:
	    insert_in_elem_constr_lists((sem_junct*)
					subconstr->child);
	    break;
      default:
	   sam_error("insert_in_elem_constr_lists: Wrong child type!\n",
		     NULL, 2);
      }
      subconstr = subconstr->next;
   }
}



void add_sem_junct(parentJunct, myJunct)
/* Add myJunct to the sub_constr_list of parentJunct. */
sem_junct*     parentJunct;
sem_junct*     myJunct;
{

  myJunct->parent_junct = parentJunct;

  if (parentJunct)
    add_to_sub_constr_list(parentJunct, 
			   (sem_child*) myJunct,
			   CHILD_JUNCT, 
			   1);
}



sem_junct*  create_sem_junct(junctType)
/* create new sem_junct on the constraint-stack and initialize it
   It is by default a root junctor.
*/
int            junctType;
{
  sem_junct* outJunct;
  
  outJunct = (sem_junct *) c_sp;
  c_sp += SIZE_IN_WORDS(sem_junct);

  /* check for stack-overflow */
  if (c_sp >= c_stack_end )
    sam_error("sem_constr.c: Constraint-Stack-Overflow", NULL, 2);
  
  outJunct->parent_junct    = NULL;
  outJunct->sub_constr_list = NULL;
  outJunct->type            = junctType;
  outJunct->arity           = 0;

  return outJunct;
}



/******* procedures for the attribute-handling in the sem_constr-tree ********/

int  remove_from_sub_constr_list(junct, child)
/* Remove child from junct's sub_constr_list 
 * return child's position in the sub_constr_list 
 */
sem_junct* junct;
sem_child* child;
{
  sem_constr_list  predecessor = NULL;
  sem_constr_list  curElement  = junct->sub_constr_list;
  int              counter     = 1;    /* child's position in list */

  if (junct->arity == 0)
    sam_error("remove_from_sub_constr_list: Arity is already 0",
	      NULL, 2);
  
  /* Find the predecessor of the element which has to be removed */
  while (curElement->child != child) {
    predecessor = curElement;
    curElement  = curElement->next;
    counter++;
    if (!curElement)
      sam_error("remove_from_sub_constr_list: child hasn't been found in sub_constr_list", 
		NULL, 2);
  }
  
  /* is child the first element of the list? */
  if (!predecessor ) {
    trail_ptr((void**) &(junct->sub_constr_list));
    junct->sub_constr_list = curElement->next;
  }
  else {
    trail_ptr((void**) &(predecessor->next));
    predecessor->next = curElement->next;
  }
  
  trail_int(&(junct->arity));
  junct->arity--;

  return counter;
}



void  remove_junct(junct)
sem_junct*  junct;
{
   sem_junct*       myParent;
   sem_junct*       newParentOfChildren;
   sem_constr_list  subconstr;
   sem_child*       child; 
   int              childType;
   int              childPos;
   int              dummy;

   if (junct->parent_junct) {
      myParent = junct->parent_junct;
      newParentOfChildren = myParent;
   }
   else {
      myParent = semantic_root;
      newParentOfChildren = NULL;
   }
   
   childPos = remove_from_sub_constr_list(myParent, (sem_child*) junct);

   subconstr = junct->sub_constr_list;
   while (subconstr) {
      child     = subconstr->child;
      childType = subconstr->child_type;
      if (childType == CHILD_ELEM) {
	 trail_ptr((void**) &(((sem_elem_constr*) child)->parent_junct));
	 ((sem_elem_constr*) child)->parent_junct = newParentOfChildren;
      }
      else {
	 trail_ptr((void**) &(((sem_junct*) child)->parent_junct));
	 ((sem_junct*) child)->parent_junct = newParentOfChildren;
      }
      dummy = remove_from_sub_constr_list(junct, child);
      add_to_sub_constr_list(myParent, child, childType, childPos);
      if (myParent != semantic_root)
	 childPos++;   /* Keep order of subconstraints */
      subconstr = subconstr->next;
   }

   trail_int(&(junct->type));
   junct->type = DEACTIVATED;
}



/* junct is a junct with arity==1.
   Removes junct if junct is not a root junctor. */
void  remove_junct_with_arity_one(junct)
sem_junct*  junct;
{
   sem_junct*        parent    = junct->parent_junct;
   sem_constr_list   subconstr = junct->sub_constr_list;
   sem_junct*        childJunct;
   
   if (parent) {
      remove_junct(junct);
      
      if (remove_cascades  &&  subconstr->child_type == CHILD_JUNCT) {
	 
	 childJunct = (sem_junct*) (subconstr->child);
	 
	 if (childJunct->type == parent->type)
	    remove_junct(childJunct);
      }
   }
}



/* Removes also root junctors. */
void  WAS_remove_junct_with_arity_one(junct)
sem_junct*  junct;
{
   sem_junct*        parent    = junct->parent_junct;
   sem_constr_list   subconstr = junct->sub_constr_list;
   sem_junct*        childJunct;

   remove_junct(junct);

   if (remove_cascades  &&  subconstr->child_type == CHILD_JUNCT) {

      childJunct = (sem_junct*) (subconstr->child);

      if (( parent  &&  childJunct->type == parent->type)  ||
	  (!parent  &&  childJunct->type == JUNCT_CONJ)) {
	 remove_junct(childJunct);
      }
   }
}



/* is_sem_junct_alive
 * return TRUE, if the constraint is neither deleted nor deactivated 
 * else      mark the whole branch of this constraint and return FALSE
 */ 
int is_sem_junct_alive(myJunct)
sem_junct*  myJunct;
{
   sem_junct*  tmpJunct;
   sem_junct*  curJunct;
   
   /* try to find a deleted or deactivated ancestor of myJunct: */ 
   tmpJunct = myJunct;
   while (tmpJunct  &&  tmpJunct->type != DELETED  
	            &&  tmpJunct->type != DEACTIVATED) {
      tmpJunct = tmpJunct->parent_junct;
   }

   /* if all ancestors of myJunct are alive, return TRUE: */
   if (tmpJunct == NULL)
      return TRUE;
   
   /* otherwise mark this branch as deleted or deactivated,
      respectively, and return FALSE: */
   curJunct = myJunct;
   while (curJunct != tmpJunct) {
      trail_int(&(curJunct->type));
      trail_int(&(curJunct->arity));
      trail_ptr((void**) &(curJunct->sub_constr_list));
      curJunct->type            = tmpJunct->type;
      curJunct->arity           = 0;
      curJunct->sub_constr_list = NULL;
      curJunct = curJunct->parent_junct;
   }
   return FALSE;
}



/* is_sem_elem_constr _alive
 * return TRUE, if the constraint is neither deleted nor deactivated 
 * else      mark the whole branch of this constraint and return FALSE
 */ 
int is_sem_elem_constr_alive(elemConstr)
sem_elem_constr*  elemConstr;
{  
   if (elemConstr->type == DEACTIVATED  ||  elemConstr->type == DELETED) 
      return FALSE;
   
   if (is_sem_junct_alive(elemConstr->parent_junct))
      return TRUE;
   
   /* if some of the ancestors of elemConstr are not alive (this means
      that elemConstr->parent_junct != NULL), set elemConstr->type to
      the value of elemConstr->parent_junct->type. This value has been
      propagated downward by is_sem_junct_alive */
   trail_int(&(elemConstr->type));  
   elemConstr->type = elemConstr->parent_junct->type;
   
   return FALSE;
}



/* Deactivate a living (alive) elemConstr, as it can't be violated any more
 * and propagate deactivation upward, i.e. deactivate disjunctions in which
 * one child is deactivated and deactivate conjunctions in which all children
 * are deactivated */
void deactivate_sem_elem_constr(elemConstr)
sem_elem_constr* elemConstr;
{
  sem_junct*  curJunct;
  sem_child*  curChild;
  int         dummy;

  /* deactivate the elementary constraint */
  trail_int(&(elemConstr->type));
  elemConstr->type = DEACTIVATED;

  /* deactivate its ancestors (if necessary) */
  curChild = (sem_child*) elemConstr;
  curJunct = elemConstr->parent_junct;

  while (curJunct) {
    
     switch (curJunct->type) {

     case JUNCT_DISJ:
	   trail_int(&(curJunct->arity));
	   trail_ptr((void**) &(curJunct->sub_constr_list));
	   curJunct->arity           = 0;
	   curJunct->sub_constr_list = NULL;
	   break;

     case JUNCT_CONJ:
	   /* Remove curChild from sub_constr_list of curJunct and
	      decrement arity of curJunct by 1: */
	   dummy = remove_from_sub_constr_list(curJunct, curChild);
	   if (curJunct->arity > 1) {
	      /* in this case deactivation doesn't have to be
		 propagated upward */ 
	      return;
	   }
	   if (curJunct->arity == 1) {
	      if (remove_arity_one)
		 remove_junct_with_arity_one(curJunct);
	      return;
	   }
	   break;

     default:
	   sam_error("deactivate_sem_elem_constr: Wrong junctor type.",
		     NULL, 2);

     } /* switch */
    
     trail_int(&(curJunct->type));
     curJunct->type = DEACTIVATED;
     
     /* go upward */
     curChild = (sem_child*) curJunct;
     curJunct = curJunct->parent_junct;
     
  } /* while */
}



/* Delete a living (alive) elemConstr, as it can't be fulfilled any more
 * and propagate deletion upward, i.e. delete conjunctions in which
 * one child is deleted and delete disjunctions in which all children
 * are deleted 
 *
 * returns FALSE, if root node has been deleted, i.e. the constraint net
 *                                                    is inconsistent 
 * returns TRUE  otherwise
 */
int  delete_sem_elem_constr(elemConstr)
sem_elem_constr*  elemConstr;
{
  sem_junct*  curJunct;
  sem_child*  curChild;
  int         oldType = elemConstr->type;
  int         dummy;

  /* delete the elementary constraint */
  trail_int(&(elemConstr->type));
  elemConstr->type = DELETED;
  
  /* delete its ancestors (if necessary) */
  curChild = (sem_child*) elemConstr;
  curJunct = elemConstr->parent_junct;

  while (curJunct) {

     switch (curJunct->type) {

     case JUNCT_DISJ:
	   /* Remove curChild from sub_constr_list of curJunct and
	      decrement arity of curJunct by 1: */
	   dummy = remove_from_sub_constr_list(curJunct, curChild);
	   if ( curJunct->arity > 1 ) {
	      /* in this case deletion doesn't have to be propagated
		 upward */
	      return TRUE;
	   }
	   if (curJunct->arity == 1) {
	      if (oldType == ELEM_EQU  &&  !try_det_inst())
		 return FALSE;
	      if (is_sem_junct_alive(curJunct)
		  &&  curJunct->arity == 1
		  &&  remove_arity_one)
		 remove_junct_with_arity_one(curJunct);
	      return TRUE;
	   }
	   break;

     case JUNCT_CONJ:
	   trail_int(&(curJunct->arity));
	   trail_ptr((void**) &(curJunct->sub_constr_list));
	   curJunct->arity           = 0;
	   curJunct->sub_constr_list = NULL;
	   break;

     default:
	   sam_error("delete_sem_elem_constr: Wrong junctor type.",
		     NULL, 2);

     } /* switch */
	   
     trail_int(&(curJunct->type));
     curJunct->type = DELETED;
     
     /* go upward */
     curChild = (sem_child*) curJunct;
     curJunct = curJunct->parent_junct;
     
  } /* while */
  
  return FALSE;
}



/* generate new elementary subconstraints from bindings, which 
 * have been written on the PU-trail during the protocol unification.
 */
void  generate_sem_elem_constraints_from_pu_trail(parentJunct, type, insertFlag)
sem_junct*  parentJunct;    /* parent junctor of new elem constrs */
int         type;   
int         insertFlag;      
{
  WORD*       left_gp;    /* left side variable of new elem constrs */
  WORD*       right_gp;   /* right side term of new elem constrs */
  s_trail*    tpt;
  sem_junct*  myJunct;    /* new junctor */

  if (pu_tr > pu_trail + 1
      &&  parentJunct  &&  parentJunct->type != JUNCT_CONJ) {
     
     /* if more than one elem constr has to be generated and
	parentJunct is not a conjunction, generate an AND junctor
	myJunct and attach myJunct to parentJunct: */
     myJunct = add_new_sem_junct(parentJunct, JUNCT_CONJ);
  }
  else {
     myJunct = parentJunct;
  }

  /***********************************************************/
  /* look in the PU-trail, which variables were bound by the */
  /* protocol-unification                                    */
  /***********************************************************/
  for (tpt = pu_tr-1; tpt >= pu_trail; tpt--) {
    
    /************************************************************/
    /* left_gp points to the T_FVAR that was bound by protocol  */
    /* unification. Look for what it's been bound to and create */
    /* a constraint                                             */
    /************************************************************/
    left_gp  = (WORD*) tpt->ptr;
    right_gp = deref(left_gp, bp);
    
    *(tpt->ptr) = tpt->oval;           /* untrail the bound var */

    add_new_sem_elem_constr(type, right_gp, myJunct, left_gp,
			    insertFlag);
  }
  
  pu_tr = pu_trail;                    /* reset pu_tr */
}



/* decompose elementary constraints which still can be fulfilled or violated
 * i.e. generate new elementary subconstraints from bindings, which 
 * have been written on the PU-trail during the protocol unification.
 * The original elementary constraint will be deactivated 
 */
void decompose_sem_elem_constr(elemConstr)
sem_elem_constr*  elemConstr;     /* elem. constr. to be decomposed */
{
   int         dummy;

   generate_sem_elem_constraints_from_pu_trail(elemConstr->parent_junct,
					       elemConstr->type,
					       TRUE); 
   
   /* deactivate elemConstr: */
   trail_int(&(elemConstr->type));
   elemConstr->type = DEACTIVATED;

   if (elemConstr->parent_junct) {
      /* Remove elemConstr from the subconstraint list of its parent junct
	 and decrement the arity of its parent junct by 1: */
      dummy = remove_from_sub_constr_list(elemConstr->parent_junct,
					  (sem_child*) elemConstr);
   }
}


/*------------------------- deterministic instantiation -----------------*/

/* Return TRUE if no disjunction with more than one subconstraint is
            found on the way up through the tree
   Return FALSE otherwise
*/
int is_sem_junct_upwards_deterministic(myJunct)
sem_junct* myJunct;
{
   while (myJunct) {
      if ((myJunct->type == JUNCT_DISJ)  &&  (myJunct->arity > 1))
	 return FALSE;
      myJunct = myJunct->parent_junct;
   }

   return TRUE;
}



/* inserts elemConstr as first element in InputList and returns the
   changed list. 
*/
sem_constr_list   insert_elem_constr(elemConstr,InputList)
sem_elem_constr*  elemConstr;
sem_constr_list   InputList;
{
   sem_constr_list   List;
   
   List = (sem_constr_list) MALLOC(sizeof(sem_constr_list_ele));
   
   List->child_type = CHILD_ELEM;
   List->child      = elemConstr;
   List->next       = InputList;
   
   return List;
}



/* Returns a list of all deterministic, living (alive) elementary
   constraints which are leafs in the tree with the root Junct.
   */
sem_constr_list   get_elem_constr_list(Junct,InputDetInstList)
sem_junct*        Junct;
sem_constr_list   InputDetInstList;
{
   sem_constr_list   DetInstList, SubConstr;
   int               ChildType;
   sem_junct*        ChildJunct;
   sem_elem_constr*  ChildElem;

   switch (Junct->type) {
      
   case JUNCT_DISJ:

	 if (Junct->arity > 1)
	    /* Junct is not downwards deterministic */
	    return InputDetInstList;
	 
	 /* Junct->arity == 1
	    i.e., Junct is downwards deterministic */

	 ChildType = Junct->sub_constr_list->child_type;

	 if (ChildType == CHILD_JUNCT) {
	    ChildJunct = (sem_junct*) Junct->sub_constr_list->child;
	    if (ChildJunct->type != DEACTIVATED  &&
		ChildJunct->type != DELETED) 
	       return get_elem_constr_list(ChildJunct, InputDetInstList);
	 }
	 else { /* ChildType == CHILD_ELEM */
	    ChildElem = (sem_elem_constr*) Junct->sub_constr_list->child;
	    if (ChildElem->type != DEACTIVATED  &&
		ChildElem->type != DELETED)
	       return insert_elem_constr(ChildElem, InputDetInstList);
	 }
	 
	 /* Child is DEACTIVATED or DELETED */
	 return InputDetInstList;
	    
   case JUNCT_CONJ:
	 
	 DetInstList = InputDetInstList;
	 SubConstr = Junct->sub_constr_list;
	 
	 while (SubConstr != NULL) {

	    ChildType = SubConstr->child_type;

	    if (ChildType == CHILD_JUNCT) {
	       ChildJunct = (sem_junct*) SubConstr->child;
	       if (ChildJunct->type != DEACTIVATED  &&
		   ChildJunct->type != DELETED)
		  DetInstList = get_elem_constr_list(ChildJunct,
						     DetInstList); 
	    }
	    else { /* ChildType == CHILD_ELEM */
	       ChildElem = (sem_elem_constr*) SubConstr->child;
	       if (ChildElem->type != DEACTIVATED  &&
		   ChildElem->type != DELETED)
		  DetInstList = insert_elem_constr(ChildElem,
						   DetInstList); 
	    }

	    SubConstr = SubConstr->next;
	 }
	 
	 return DetInstList;
      
   case DEACTIVATED:
   case DELETED:
	 return InputDetInstList;
	 
   default: sam_error("get_elem_constr_list: Wrong junctor type.",
		      NULL, 2);
	 
   }

   /* gcc doesn't recognize that every case has a return. The
      following return has been added to avoid a gcc warning: */
   return InputDetInstList;
}



void  free_det_inst_list(list)
sem_constr_list  list ;
{
   sem_constr_list  next;

   while (list != NULL) {
      next = list->next;
      FREE((void*) list, sizeof(sem_constr_list_ele));
      list = next;
   }
}



/* Tries deterministic instantiation for the elementary constraint
   myElem:
   returns FALSE if deterministic instantiation failed
   returns TRUE otherwise, i.e., if deterministic instantiation was
             successful or if deterministic instantiation could not be
	     tried because myElem was not alive
 */
#define FAIL_U       {return FALSE;}
#define UNIF_FAIL(x) { \
switch (solution_computation_mode) { \
case NO_SOL_CMP: unif_fails ++;         break; \
case LOCAL_CHK:  loc_cons_unif_fails++; break; \
case GLOBAL_CHK: gl_cons_unif_fails++;  break; \
case INST_TREES: inst_anl_unif_fails++; break; \
default: \
      sam_error("undefined solution computation mode", NULL, 2); \
} \
FAIL_U}


int try_det_inst_elem(myElem)
sem_elem_constr* myElem;
{
   WORD*            derefLeftSide;
   WORD*            derefRightSide;
   /*
   WORD*            checkVar;
   int              setConsFailOccFlag;
   */
   
   if (is_sem_elem_constr_alive(myElem)) {
      
      switch (myElem->type) {
	 
      case ELEM_EQU:

	    derefLeftSide  = deref(myElem->var, bp);
	    derefRightSide = deref(myElem->right_side, bp);

	    /* Decide whether or not consistency_fail_occurred flag
	       has to be set after successful UNIFY: */
	    if (!chk_arg(derefLeftSide, bp, derefRightSide, bp)) {

	       /* Computation of setConsFailOccFlag leads to
		  INCOMPLETENESS !!!
	       if (!chp->consistency_fail_occurred) {
		  if (GETTAG(*derefLeftSide) == T_FVAR) {
		     if (GETTAG(*derefRightSide) == T_FVAR)
			checkVar = MAXIMUM(derefLeftSide,
					   derefRightSide);
		     else
			checkVar = derefLeftSide;
		  }
		  else {
		     if (GETTAG(*derefRightSide) == T_FVAR)
			checkVar = derefRightSide;
		     else
			checkVar = NULL;
		  }
		  
		  if (!checkVar)
		     setConsFailOccFlag = TRUE;
		  else
		     setConsFailOccFlag = is_variable_relevant(checkVar);
	       }
	       else {
		  setConsFailOccFlag = FALSE;
	       }
	       */


	       /* try to make the deterministic instantiation.
		  UNIFY calls macros like UN_FVAR, and UN_FVAR calls
		  FAIL_U or UNIF_FAIL if not successful. */
	       UNIFY(derefLeftSide, bp, derefRightSide, bp);

	       /**********************************************/
	       /* the following lines of try_det_inst_elem   */
	       /* are only processed if UNIFY was successful */
	       /**********************************************/

	       /* Computation of setConsFailOccFlag leads to
		  INCOMPLETENESS !!!
		  if (setConsFailOccFlag) */
	       chp->consistency_fail_occurred = TRUE;
	       
	    }
	       
	    if (is_sem_elem_constr_alive(myElem))
		deactivate_sem_elem_constr(myElem);

	    return TRUE;
	    
      case ELEM_DIFF:

	    sam_error("try_det_inst_elem: deterministic instantiation tried for disequality constraints", NULL, 2);

	    pu_tr = pu_trail;
	    if (un_protoco(myElem->var, bp, myElem->right_side, bp, 1)) {
	       /* any unifications made? */
	       if (pu_tr == pu_trail) {
		  /* var and right_side are equal
		   * delete myElem, for it cannot be fulfilled any more 
		   * if the root-node is deleted, a fail has to follow
		   */
		  return delete_sem_elem_constr(myElem);
	       }
	       else {
		  /* decompose the constraint net with the
		     unifications on the pu_trail */ 
		  decompose_sem_elem_constr(myElem);
		  return TRUE;
	       }
	    } /* if unprotoco */
	    else {
	       /* no unifications are possible, therefore the
		  constraint can't be violated any more */
	       pu_untrail(); 
	       deactivate_sem_elem_constr(myElem);
	       return TRUE;
	    }
      } /* switch */
   }
   return TRUE;
}

#undef FAIL_U
#undef UNIF_FAIL



/* Tries deterministic instantiation at a junctor Junct:
   returns FALSE if deterministic instantiation was not successful
   returns TRUE otherwise, i.e., if deterministic instantiation was
             successful or if deterministic instantiation has not been
	     tried because Junct was not upwards deterministic
 */
int  try_det_inst() {
#define DEFINE_try_det_inst
#ifdef DEFINE_try_det_inst
   sem_elem_constr*  ElemConstr;
   sem_constr_list   ElemConstrListPtr, ElemConstrList;
   WORD*             c_sp_save;

   do {
      c_sp_save = c_sp;
      ElemConstrList = get_elem_constr_list(semantic_root, NULL); 
      ElemConstrListPtr = ElemConstrList;
#ifdef STATISTICS1
      if (ElemConstrListPtr) {
	 det_inst_tries++;
	 switch (solution_computation_mode) {
	 case NO_SOL_CMP:  other_det_inst_tries++;     break;
	 case LOCAL_CHK:   loc_cons_det_inst_tries++;  break;
	 case GLOBAL_CHK:  gl_cons_det_inst_tries++;   break;
	 case INST_TREES:  inst_anl_det_inst_tries++;  break;
	 default: sam_error("undefined solution computation mode",  
			    NULL, 2); 
	 }
      }
#endif
      
      while (ElemConstrListPtr != NULL) {
	 ElemConstr = (sem_elem_constr*) ElemConstrListPtr->child;
	 if (try_det_inst_elem(ElemConstr) == 0) { 
	    free_det_inst_list(ElemConstrList);
#ifdef STATISTICS1
	    det_inst_fails++; 
#endif
	    return FALSE;
	 }
	 ElemConstrListPtr = ElemConstrListPtr->next;
      }
      
      free_det_inst_list(ElemConstrList);
      
   } while (c_sp_save != c_sp);
   /* If c_sp_save>c_sp, during try_det_inst_elem new sem elem
      constrs might have been generated (by
      decompose_sem_elem_constr). Of cource, these have also to be
      processed. */
   
#endif
   return TRUE;
}
