/******************************************************/
/*    S E T H E O                                     */
/*                                                    */
/* FILE: sem_tools.c                                  */
/* VERSION:                                           */
/* DATE:                                              */
/* AUTHOR:                                            */
/* NAME OF FILE:                                      */
/* DESCR:                                             */
/* MOD:                                               */
/* BUGS:                                              */
/******************************************************/

/* Provides helper functions for semantic constraint
   handling (some of them may be obsolete by now):
   - i_equconstr.c
   - i_disconstr.c
   - i_conjconstr.c
   - sem_constr.h
   - sem_consistency.h
   */

#include <stdio.h>
#include "tags.h"
#include "machine.h"
#include "antilemma.h"
#include "deref.h"
#include "errors.h"
#include "sem_constr.h"
#include "sem_consistency.h"
#include "sem_tools.h"
#include "symbtab.h"
#include "termcomplexity.h"
#include "unitlemma.h"
#include "univ_constr.h"
#define INSTR_LENGTH		2



/*********** procedures for list handling **************/

int is_list(list)
WORD* list;
{
   return (ISSYMBOL(*list) 
	   && (GETSYMBOL(*list) == LISTFUNCT
	       || GETSYMBOL(*list) == EMPTYLIST));
}



void check_list(list)
/* Is list a list? If not, signal an error */
WORD* list;
{
   if (!is_list(list))
      sam_error("check_list: list expected", list, 3);
}



void check_nonempty_list(list)
/* Is list a non-empty list? If not, signal an error */
WORD* list;
{
   if (!(ISSYMBOL(*list)  &&  GETSYMBOL(*list) == LISTFUNCT))
      sam_error("check_nonempty_list: non-empty list expected", 
		list, 3);
}



int is_list_empty(list)
WORD* list;
{
   if (!is_list(list))
      sam_error("is_list_empty: argument not a list", list, 3);
   return (ISSYMBOL(*list)  &&  GETSYMBOL(*list) == EMPTYLIST);
}



int list_length(list) 
WORD* list;
{
   int length = 0;
   while (!is_list_empty(list)) {
      length++;
      list = deref(list+2, bp); /* cdr */
   }
   return length;
}



sem_constr_list  push(junct, list)
sem_junct*       junct;
sem_constr_list  list;
{
   sem_constr_list  newElement;

   newElement = (sem_constr_list) MALLOC(sizeof(sem_constr_list_ele));

   newElement->child      = junct;
   newElement->child_type = CHILD_JUNCT;
   newElement->next       = list;

   return newElement;
}



sem_constr_list  pop(list)
sem_constr_list  list;
{
   sem_constr_list  nextElement;
   
   nextElement = list->next;
   FREE(list, sizeof(sem_constr_list_ele));

   return nextElement;
}



int  member(junct, list)
sem_junct*       junct;
sem_constr_list  list;
{
   sem_constr_list  element;

   element = list;
   while (element) {
      if (((sem_child*) junct) == element->child) {
	 return TRUE;
      }
      element = element->next;
   }

   return FALSE;
}



/*********** procedures for constraint input **************/

void extract_constraint_sides(constraint, leftListPtr, rightListPtr)
/* constraint must be a list containing (exactly) two non-empty lists,
 * if not: error! Extract these and put 'em in leftListPtr and
 rightListPtr */ 
WORD*  constraint;
WORD** leftListPtr;
WORD** rightListPtr;
{
   check_nonempty_list(constraint);         /* header correct? */
   *leftListPtr = deref(constraint+1, bp);  /* car of constraint (first list element) */
   check_nonempty_list(*leftListPtr);

   constraint = deref(constraint+2, bp);    /* cdr of constraint */
   check_nonempty_list(constraint);         /* header correct? */
   *rightListPtr = deref(constraint+1, bp); /* car (second list element) */
   check_nonempty_list(*rightListPtr);
   
   /* constraint must be empty now (i.e. there must be exactly two
      elements in the list) */ 
   constraint = deref(constraint+2, bp);    /* cdr of constraint */
   if ( !(ISSYMBOL(*constraint) && (GETSYMBOL(*constraint) == EMPTYLIST)) )
      sam_error("extract_constraint_sides: end-of-list expected",
		constraint, 3); 
}



/*********** procedures for constraint and variable output **************/

void print_indentation(indent)
int indent;
{
  printf("\n");
  while(indent > 0) {
    printf("    ");
    indent --;
  }
}



void print_sem_elem_constr(self, indent)
sem_elem_constr* self;
int              indent;
{
  sem_elem_constr* curConstr = NULL;

  switch(self->type) {

  case ELEM_EQU:
    print_indentation(indent);
    disp_(stdout, self->var, bp);
    printf(" == ");
    disp_(stdout, self->right_side, bp); 
    fflush(stdout);
    /* DEBUGGING */
    /* curConstr = ( (constr_block*)GETPTR(*(self->var)) )->sem_equ_constr; */
    break;

  case ELEM_DIFF:
    print_indentation(indent);
    disp_(stdout, self->var, bp);
    printf(" =/= ");
    disp_(stdout, self->right_side, bp);
    fflush(stdout);
    curConstr = ( (constr_block*)GETPTR(*(self->var)) )->sem_diff_constr;
    break;

  case DELETED:
    /* in this case this part of the tree shouldn't be displayed */
    print_indentation(indent);
    printf("DELETED elem constr");
    fflush(stdout);
    break;

  case DEACTIVATED:
    /* in this case this part of the tree shouldn't be displayed */
    print_indentation(indent);
    printf("DEACTIVATED elem constr");
    fflush(stdout);
    break;

  default:
    sam_error("Wrong type in elementary subconstraint.", NULL, 2);
  }

#warning self-information disabled
  /*
  printf(" (self: "GETVALFORMAT", parent: "GETVALFORMAT", next: "GETVALFORMAT")", 
	 (WORD_cast) self, 
	 (WORD_cast) self->parent_junct, 
	 (WORD_cast) self->next);
  fflush(stdout);
  */

  /* DEBUGGING */
  return;
  
  /* Does the variable have the correct back-pointer to the constraint? */
  if (self->type == ELEM_EQU  ||  self->type == ELEM_DIFF) {
     if (GETTAG(*(self->var)) != T_FVAR) {
	sam_error("print_sem_elem_constr: back-pointer to bound variable.",
		  self->var, 1);
	return;
     }
     while(curConstr && (curConstr != self)) {
	curConstr = curConstr->next;
     }
     if (!curConstr)
	sam_error("print_sem_elem_constr: no pointer from variable to current constraint!", 
		  NULL, 2);
  }
}



void print_sem_constr_tree(self, indent)
/* Display junctor self and its sons with the indentation indent */
sem_junct* self;
int        indent;
{
  sem_constr_list curEle = self->sub_constr_list;

  switch(self->type) {

  case JUNCT_DISJ:
    print_indentation(indent);
    printf("OR");
    fflush(stdout);
    break;

  case JUNCT_CONJ:
    print_indentation(indent);
    printf("AND");
    fflush(stdout);
    break;

  case DELETED:
    /* in this case this part of the tree shouldn't be displayed */
    print_indentation(indent);
    printf("DELETED junct");
    fflush(stdout);
    break;

  case DEACTIVATED:
    /* in this case this part of the tree shouldn't be displayed */
    print_indentation(indent);
    printf("DEACTIVATED junct");
    fflush(stdout);
    break;

  default:
    print_indentation(indent);
    printf("junctor type: %d\n", self->type);
    fflush(stdout);
    sam_error("Wrong junctor type.", NULL, 2);
  }

#warning self-information disabled
  /*
  printf(" (self: "GETVALFORMAT", type: %i, arity: %i, parent: "GETVALFORMAT")",
	 (WORD_cast) self,
	 self->type,
	 self->arity,
	 (WORD_cast) self->parent_junct);
  fflush(stdout);
  */
  
  if (self->type == JUNCT_CONJ  ||  self->type == JUNCT_DISJ) {
     while(curEle) {
	switch(curEle->child_type) {
	case CHILD_JUNCT:
	      print_sem_constr_tree((sem_junct*)curEle->child,
				    indent+1); 
	      fflush(stdout);
	      break;
	case CHILD_ELEM:
	      print_sem_elem_constr((sem_elem_constr*)curEle->child,
				    indent+1); 
	      break;
	default:
	      print_indentation(indent);
	      printf("Wrong tag in sub_constr_list: %i",
		     curEle->child_type); 
	      fflush(stdout);
	}
	curEle = curEle->next;
     }
  }
}
   


/* Display all semantic constraint trees which have an elementary constraint
   with left side variable var */
void  print_sem_constr_trees_at_variable(var)
WORD*  var;
{
   sem_constr_list   root_list,root_list_ele;

   printf("print_sem_constr_trees_at_variable(");
   disp_(stdout,var,bp);
   printf("):\n");
   fflush(stdout);

   if (GETTAG(*var) != T_FVAR) {
      printf("Variable is already bound.\n");
      fflush(stdout);
      return;
   }

   root_list = compute_list_of_root_nodes_at_variable(var); 
   
   /* Print trees of computed root nodes and free root_list: */

   if (root_list == NULL) {
      printf("Variable has no living (alive) constraints.\n");
      fflush(stdout);
      return;
   }

   while (root_list != NULL) {
      root_list_ele = root_list;
      print_sem_constr_tree((sem_junct*) root_list_ele->child, 0);
      root_list = root_list->next;
      FREE(root_list_ele, sizeof(sem_constr_list_ele));
   }
}



/* Display all syntactic constraints with left side var. */
void  print_syn_constrs_of_variable(var)
WORD*  var;
{
   WORD*          myVar      = deref(var, bp);
   constr_right*  synConstrs = ((constr_block*) GETPTR(*myVar))->syn_constr;

   printf("printing syntactic constraints of ");
   disp_(stdout, myVar, bp);
   printf(":");
   /* printf("\n"); */
   fflush(stdout);

   while (synConstrs) {
      print_indentation(1);
      disp_(stdout, myVar, bp);
      printf(" =/= ");
      disp_con(deref(synConstrs->sub_gp, bp), bp);
      printf(" (type = %d, counter = %d)",
	     GETTAG(synConstrs->zaehler->next),
	     synConstrs->zaehler->length);
      fflush(stdout);
      
      synConstrs = synConstrs->next;
   }

   printf("\n");
   fflush(stdout);
}



/* Display the syntactic constraints of all variables of chpt. */
void  print_syn_constrs_of_chpt(chpt)
choice_pt*  chpt;
{
   int    i; 
   WORD*  var;

   printf("Printing syn constrs of variables at choice point %x:\n",
	  chpt); 
   for ( i = 0 ; i < chpt->nr_fvars ; i++ ) {
      if (GETTAG(*(chpt->block_fvars + i)) == T_BVAR) {
	 var = GETPTR(*(chpt->block_fvars + i));
	 print_syn_constrs_of_variable(var);
      }
   }
}



void  print_instantiations(varBlock, nrVars, GETPTR_flag)
WORD*  varBlock;
int    nrVars;
int    GETPTR_flag;
{
   int               i;
   WORD*             var;
   sem_elem_constr*  constr;

   printf("printing %d instantiations:\n", nrVars);

   for ( i=0 ; i<nrVars ; i++ ) {
      if (GETPTR_flag)
	 var = deref(GETPTR(*(varBlock+i)), bp);
      else
	 var = deref(varBlock+i, bp);
      
      printf("%d: ", i);
      disp_(stdout, var, bp);
      if (GETTAG(*var) == T_FVAR) {
	 constr = ((constr_block*) GETPTR(*var))->sem_equ_constr;
	 while (constr) {
	    if (is_sem_elem_constr_alive(constr)) {
	       printf("   (with alive elem equ constrs)");
	       break;
	    }
	    constr = constr->next;
	 }
      }
      printf("\n");
      fflush(stdout);
   }
}



/***** Helper functions for constraint list handling *****/

sem_constr_list add_to_junct_list(junctList, junct)
sem_constr_list junctList;
sem_junct*      junct;
{
   sem_constr_list   newEle;
   
   newEle = (sem_constr_list) MALLOC(sizeof(sem_constr_list_ele));
   newEle->child = (void*) junct;
   newEle->next  = junctList;

   return newEle;
}



void free_sem_constr_list(list)
sem_constr_list list;
{
   sem_constr_list   prevElement;
   
   while(list) {
      prevElement = list;
      list = list->next;
      FREE(prevElement, sizeof(sem_constr_list_ele));
   }
}



int junct_list_length(junctList)
sem_constr_list junctList;
{
   int count = 0;

   while(junctList) {
      count++;
      junctList = junctList->next;
   }

   return count;
}



int junct_list_contains(junctList, junct)
sem_constr_list junctList;
sem_junct*      junct;
{
   while(junctList) {
      if (junctList->child == (void*)junct) return TRUE;
      junctList = junctList->next;
   }
   return FALSE;
}



/******************* More helper functions *******************/

/* Find the root junct of a given (elementary) constraint
   Helper function */
sem_junct* get_root_junct(constr)
sem_elem_constr* constr;
{
   sem_junct*   curJunct = constr->parent_junct;
   
   if (!curJunct) 
      return NULL;

   while(curJunct->parent_junct)
      curJunct = curJunct->parent_junct;

   return curJunct;
}



sem_constr_list  compute_list_of_root_nodes_at_variable(var)
WORD*  var;
{
   sem_elem_constr*  eq_constr;
   sem_junct*        constr;
   sem_child*        child;
   int               child_type;
   sem_constr_list   root_list,root_list_ele,list_ptr;
   int               is_member;

   root_list = NULL;

   /* Get first elementary equality constraint to check: */
   eq_constr = ((constr_block*) GETPTR(*var))->sem_equ_constr;
   while (eq_constr != NULL) {
      if (!is_sem_elem_constr_alive(eq_constr))
	 eq_constr = eq_constr->next;
      else
	 break;
   }
   
   /* Process all elementary equality constraints of var: */
   while (eq_constr != NULL) {
	 
      /* Get root node belonging to eq_constr: */
      if (eq_constr->parent_junct) {
	 constr = eq_constr->parent_junct;
	 while (constr->parent_junct != NULL)
	    constr = constr->parent_junct;
	 child = (sem_child*) constr;
	 child_type = CHILD_JUNCT;
      }
      else {
	 child = (sem_child*) eq_constr;
	 child_type = CHILD_ELEM;
      }
	 
	 /* Check whether found root node is already element of root_list: */
	 is_member = FALSE;
	 list_ptr = root_list;
	 while (list_ptr != NULL  &&  !is_member) {
	    if (list_ptr->child == child)
	       is_member = TRUE;
	    else
	       list_ptr = list_ptr->next;
	 }
	 

      /* If found root node is not yet element of root_list, insert it
	 in root_list: */
      if (!is_member) {
	 root_list_ele = (sem_constr_list) MALLOC(sizeof(sem_constr_list_ele));
	 root_list_ele->child = child;
	 root_list_ele->child_type = child_type;
	 root_list_ele->next = root_list;
	 root_list = root_list_ele;
      }
      
      /* Get next elementary equality constraint to check: */
      eq_constr = eq_constr->next;
      while (eq_constr != NULL) {
	 if (!is_sem_elem_constr_alive(eq_constr))
	    eq_constr = eq_constr->next;
	 else
	    break;
      }
   }

   return root_list;
}



WORD  *copy_complex_term(ga)
WORD       *ga;
{
   int               arity;
   WORD              *memory,*gai,*term;
   register int      i;
   
   arity = GETARITY(*ga);
   memory = sp;
   sp += arity+2;

   if (sp >= up_stack )
      sam_error("Stack-Overflow", NULL, 2);
   
   /* Put function symbol and T_EOSTR tag in the memory: */
   GENSYMBOL(*memory, GETSYMBOL(*ga), T_CRTERM);
   GENOBJ(*(memory+arity+1), 0, T_EOSTR);
   
   /* Handle the arguments: */
   for ( i=1 ; i<=arity ; i++ ) {
      
      gai = deref(ga+i, bp);

      switch (GETTAG(*gai)) {
	 
      case T_CONST:  
	    *(memory+i) = *gai;
	    break;
	    
      case T_FVAR:  
      case T_GTERM:  
	    GENPTR(*(memory+i),gai); 
	    break;
	    
      case T_NGTERM:
      case T_CRTERM: 
	    term = copy_complex_term(gai);
	    if (term == NULL)
	       return(NULL);
	    GENPTR(*(memory+i),term); 
	    break;
	    
      default:       
	    sam_error("copy_complex_term: illegal tag",
		      gai, 3);
      }
   }
   
   return(memory);
}



WORD  *copy_instantiations(choicePoint)
choice_pt  *choicePoint;
{
   int              arity;
   WORD             *memory,*gai,*term,*varBlock;
   register int     i;

   arity = choicePoint->nr_fvars;
   varBlock = choicePoint->block_fvars;

   memory = sp;
   sp += arity;
   
   if (sp >= up_stack )
      sam_error("Stack-Overflow", NULL, 2);

   /* Handle the variables: */
   for ( i=0 ; i<arity ; i++ ) {
      
      gai = deref(GETPTR(*(varBlock+i)), bp);
      
      switch (GETTAG(*gai)) {
	    
      case T_CONST:  
	    *(memory+i) = *gai;
	    break;
	 
      case T_FVAR:   
      case T_GTERM:  
	    GENPTR(*(memory+i),gai);
	    break;
	    
      case T_NGTERM:
      case T_CRTERM: 
	    term = copy_complex_term(gai);
	    if (term == NULL)
	       return(NULL);
	    GENPTR(*(memory+i),term);
	    break;
	    
      default:       
	    sam_error("copy_instantiations: illegal tag",
		      gai, 3);
      }
   }
   
   return(memory);
}



int  is_variable_relevant(var)
WORD*   var;
{
   WORD*  baseAddress  = chp->block_fvars;
   int    infimum      = -1;
   int    suprenum     = chp->nr_fvars;
   int    middle;
   WORD*  curPtr;
   WORD*  curVar;

   while (infimum + 1 < suprenum) {
      middle = (infimum + suprenum) / 2;
      curPtr = baseAddress + middle;
      curVar = GETPTR(*curPtr);
      if (curVar == var) {  /* var has been found */
	 if (GETTAG(*curPtr) == T_BVAR)
	    return TRUE;
	 else  /* var is marked as irrelevant */
	    return FALSE;
      }
      else {
	 if (curVar < var)
	    suprenum = middle;
	 else  /* curVar > var */
	    infimum = middle;
      }
   }

   /* If var does not occur in chp->block_fvars, it cannot be
      relevant: */
   return FALSE;
}



/************* procedures for constraint evaluation *************/

int  dag_tc_of_sem_elem_constr(constr, limit)
sem_elem_constr*  constr;
int               limit;
{
   s_trail*  saved_tr = tr;
   int       retval = dag_termcomplexity(constr->right_side, 1, bp,
					 limit); 
   untrail(saved_tr);
   tr = saved_tr;

   return retval;
}



int nr_of_top_level_constraints_satisfied_by_elem_constr(constr)
sem_elem_constr* constr;
{
   sem_constr_list localJunctList = NULL;
   sem_junct*      curRootJunct;
   int             retVal;
   
   /* Go to the first constraint of the variable of constr */
   constr = ((constr_block*) GETPTR(*(constr->var)))->sem_equ_constr;
   
   while(constr) {
      curRootJunct = get_root_junct(constr);
      /* add this root junct to the list, if
       * 1. constr has a root junct, i.e.curRootJunct != NULL 
       * 2. curJunct isn't a member of our list already */
      if (curRootJunct  &&  !junct_list_contains(localJunctList, curRootJunct))
	 localJunctList = add_to_junct_list(localJunctList, curRootJunct);
      constr = constr->next;
   }
   
   retVal = junct_list_length(localJunctList);
   free_sem_constr_list(localJunctList);
   return retVal;
}



/***** procedures for variable evaluation *****/

int  nr_of_constraints_at_variable(myVar)
WORD*  myVar;
{
   sem_elem_constr*  myConstr = ((constr_block*)
				 GETPTR(*myVar))->sem_equ_constr; 
   int               count    = 0;
   
   while (myConstr) {
      if (is_sem_elem_constr_alive(myConstr))
	 count++;
      myConstr = myConstr->next;
   }

   return count;
}



int  nr_of_top_level_constraints_at_variable(var)
WORD*  var;
{
   sem_elem_constr  *elemConstr;
   sem_junct        *junct;
   s_trail*         saveTr;
   int              counter = 0;
   sem_constr_list  constrListEle;

   if (GETTAG(*var) != T_FVAR)
      sam_error("nr_of_top_level_constraints_at_variable: Variable is already bound.", var, 3);
   
   /* Save trail pointer: */
   saveTr = tr;
   
   /* Get root node of all elem constrs of var and mark root nodes: */
   elemConstr = ((constr_block*) GETPTR(*var))->sem_equ_constr;
   while (elemConstr) {
      junct = elemConstr->parent_junct;
      if (!junct) {
	 trail_int(&(elemConstr->type));
	 elemConstr->type = MARKED_CONSTR;
      }
      else {
	 while (junct->parent_junct)
	    junct = junct->parent_junct;
	 trail_int(&(junct->type));
	 junct->type = MARKED_CONSTR;
      }
      elemConstr = elemConstr->next;
   }

   /* Count marked root nodes: */
   constrListEle = semantic_root->sub_constr_list;
   while (constrListEle) {
      switch (constrListEle->child_type) {
      case CHILD_ELEM:
	    if (((sem_elem_constr*) constrListEle->child)->type 
		== MARKED_CONSTR)
	       counter++;
	    break;
      case CHILD_JUNCT:
	    if (((sem_junct*) constrListEle->child)->type 
		== MARKED_CONSTR)
	       counter++;
	    break;
      default:
	    sam_error("nr_of_top_level_constraints_at_variable: Illegal child type.", NULL, 2);
      }
      constrListEle = constrListEle->next;
   }

   /* Untrail the markings: */
   untrail(saveTr);
   tr = saveTr;

   return counter;
}



int  max_nr_of_appearances_in_same_parent_AND_node(var)
WORD*  var;
{
   s_trail*          saveTrail          = tr;
   int               maxNrOfAppearances = 0;
   sem_elem_constr*  constr;

   /* Set the arity of all parent junctors to zero: */
   constr = ((constr_block*) GETPTR(*var))->sem_equ_constr;
   while (constr) {
      if (is_sem_elem_constr_alive(constr)
	  &&  constr->parent_junct->type == JUNCT_CONJ) {
	 trail_int(&(constr->parent_junct->arity));
	 constr->parent_junct->arity = 0;
      }
      constr = constr->next;
   }

   /* If constr is an elementary equality constraint and its parent
      junctor is a conjunction, increase its parent junct's arity by
      one. Set maxNrOfAppearances to 
      max(maxNrOfAppearances, parent junct's arity): */
   constr = ((constr_block*) GETPTR(*var))->sem_equ_constr;
   while (constr) {
      if (constr->type == ELEM_EQU
	  &&  constr->parent_junct->type == JUNCT_CONJ) {
	 constr->parent_junct->arity++;
	 if (constr->parent_junct->arity > maxNrOfAppearances) {
	    maxNrOfAppearances = constr->parent_junct->arity;
	 }
      }
      constr = constr->next;
   }

   /* Reset the arity of all parent junctors: */
   untrail(saveTrail);
   tr = saveTrail;

   return maxNrOfAppearances;
}



/* Returns the number of appearances of elementary constraints
   (regardless whether equality or disequality) with left side
   variable var in junct (regardless whether conjunction or
   disjunction constraints). */
int  nr_of_appearances(var, junct)
WORD*       var;
sem_junct*  junct;
{
   int               nrOfAppearances = 0;
   sem_constr_list   childListElem;
   sem_elem_constr*  elemConstr;

   childListElem = junct->sub_constr_list;
   while (childListElem) {

      if (childListElem->child_type == CHILD_ELEM) {

	 elemConstr = (sem_elem_constr*) childListElem->child;
	 if (is_sem_elem_constr_alive(elemConstr)) {

	    switch (GETTAG(*(elemConstr->var))) {
	    case T_FVAR:
		  if (elemConstr->var == var)
		     nrOfAppearances++;
		  break;
	    default:
		  sam_error("nr_of_appearances: wrong tag",
			    elemConstr->var, 3);
	    }
	 }
      }
      childListElem = childListElem->next;
   }

   return nrOfAppearances;
}



float  AverageDagTC_of_SemEqConstrsAtVariable(var)
   WORD*  var;
{
   sem_elem_constr*  constr; 
   int               nr_of_sem_elem_constrs = 0;
   int               sum_of_dag_tcs         = 0;
   
   if (GETTAG(*var) != T_FVAR)
      sam_error("AverageDagTC_of_SemEqConstrsAtVariable: Variable is already bound.",
		var, 3);
   
   constr = ((constr_block*) GETPTR(*var))->sem_equ_constr; 

   if (constr == NULL)
      return 0.0;

   while (constr != NULL) {
      if (is_sem_elem_constr_alive(constr)) {
	 nr_of_sem_elem_constrs++;
	 sum_of_dag_tcs += dag_tc_of_sem_elem_constr(constr,
						     MAXNUMBER);
      }
      constr = constr->next;
   }
   
   return ((float) sum_of_dag_tcs / (float) nr_of_sem_elem_constrs);
}



/***** procedures for input administration *****/

void  attach_to_imagined_root_node(junct)
sem_junct*  junct;
{
   /* Set flag in order to indicate the existence of semantic
      constraints: */
   if (!sem_constrs_in_proof) {
      trail_int(&sem_constrs_in_proof);
      sem_constrs_in_proof = TRUE;
      sem_constrs_during_search = TRUE;
   }

   bp->nr_constr_trees++;  /* nr of children of semantic_root which
			      have been added with current clause
			      (used for antilemmata) */
   
   /* insert junct in subconstr list of imagined root node
      (but make sure that junct->parent_junct does not point
      to semantic_root): */
   add_to_sub_constr_list(semantic_root, 
			  (sem_child*) junct,
			  CHILD_JUNCT,
			  1);
}



void  adapt_local_consistency_check()
{
/*
   adapt_rate = (adapt_rate + (  (float) adapt_gl_cons_fails 
			       / (float) adapt_gl_cons_chks )) / 2.0;
*/
   adapt_rate = (  (float) adapt_gl_cons_fails 
                 / (float) adapt_gl_cons_chks);
   
   /* If rate of failed global consistency checks is too low: */
   if (adapt_rate * 100.0 < (float) adapt_limit / 2.0) {
      check_frequency = MINIMUM(check_frequency + 1,
				constr_tree_count);
   }
   
   /* If rate of failed global consistency checks is too high: */
   if (adapt_rate * 100.0 > (float) adapt_limit) {
      check_frequency = MAXIMUM(check_frequency - 1,
				1);
   }
   
   min_check_frequency = MINIMUM(min_check_frequency,
				 check_frequency);
   max_check_frequency = MAXIMUM(max_check_frequency,
				 check_frequency);
   
   adapt_gl_cons_chks = adapt_gl_cons_fails = 0;
}



/* check local consistency at the nrTreesToCheck newest alive
   constraint trees: */
int  check_local_consistency(chpt, nrTreesToCheck)
choice_pt*  chpt;
int         nrTreesToCheck;
{
   environ*  chptHead_bp = chpt->head_bp;
   WORD*     vars_1;
   int       nrVars_1;
   int       deleted_flag_1;
#ifdef DEFINE_1_consistency_check_of_all_vars
   WORD*     vars_2;
   int       nrVars_2;
   int       deleted_flag_2;
#endif

   loc_cons_chks++;
   if (solution_computation_mode != NO_SOL_CMP)
      sam_error("Wrong solution computation mode.", NULL, 2);
   solution_computation_mode = LOCAL_CHK;

#warning one_consistency_check disabled !!!!!!!!!!!!!!!!
   /*
   one_cons_chks++;
   if (chptHead_bp) {
      vars_1   = (WORD*) chptHead_bp + SIZE_IN_WORDS(environ);
      nrVars_1 = (GETNUMBER(chptHead_bp->lcbptr->nr_vars)
		  - GETNUMBER(chptHead_bp->lcbptr->nr_strvars));
      if (!one_consistency_check(vars_1, nrVars_1, FALSE, &deleted_flag_1)) {
	 one_cons_fails++;
	 if (solution_computation_mode != LOCAL_CHK)
	    sam_error("Wrong solution computation mode.", NULL, 2);
	 solution_computation_mode = NO_SOL_CMP;
	 return FALSE;
      }
   }
   */
#ifdef DEFINE_1_consistency_check_of_all_vars
   vars_2   = chpt->block_fvars;
   nrVars_2 = chpt->nr_fvars;
   if (!one_consistency_check(vars_2, nrVars_2, TRUE, &deleted_flag_2)) {
      one_cons_fails++;
      if (solution_computation_mode != LOCAL_CHK)
	 sam_error("Wrong solution computation mode.", NULL, 2);
      solution_computation_mode = NO_SOL_CMP;
      return FALSE;
   }
   if (chptHead_bp) {
      while (deleted_flag_2) {
	 if (!one_consistency_check(vars_1, nrVars_1, FALSE, &deleted_flag_1)) {
	    one_cons_fails++;
	    if (solution_computation_mode != LOCAL_CHK)
	       sam_error("Wrong solution computation mode.", NULL, 2);
	    solution_computation_mode = NO_SOL_CMP;
	    return FALSE;
	 }
	 if (deleted_flag_1) {
	    if (!one_consistency_check(vars_2, nrVars_2, TRUE, &deleted_flag_2)) {
	       one_cons_fails++;
	       if (solution_computation_mode != LOCAL_CHK)
		  sam_error("Wrong solution computation mode.", NULL, 2);
	       solution_computation_mode = NO_SOL_CMP;
	       return FALSE;
	    }
	 }
      }
   }
#endif


   if (nrTreesToCheck != 0
       &&  ((selective_loc_cons_chk
	     &&  sem_constr_added
	     &&  constr_tree_count % check_frequency == 0)
	    ||  (!selective_loc_cons_chk
		 &&  total_inf % check_frequency == 0))
       &&  !try_to_instantiate_trees(chpt, nrTreesToCheck)) {

      sem_constr_added = FALSE;
      if (solution_computation_mode != LOCAL_CHK)
	 sam_error("Wrong solution computation mode.", NULL, 2);
      solution_computation_mode = NO_SOL_CMP;
      return FALSE;
   }

   sem_constr_added = FALSE;
   if (solution_computation_mode != LOCAL_CHK)
      sam_error("Wrong solution computation mode.", NULL, 2);
   solution_computation_mode = NO_SOL_CMP;
   return TRUE;
}



/***** procedures for lcanl/gcanl generation *****/

a_list*  generate_canl(historyForCanl, history, canl_list)
history_list  historyForCanl;
history_list  history;
a_list*       canl_list;
{
   a_list*         canl_memory;
   WORD*           canl_memory_left;
   WORD*           canl_memory_right;
   int             arity;
   WORD*           next;
   history_list    tmpHistory;
   int             i;
   WORD*           rightSide;
   WORD*           term;

   if (!historyForCanl  ||  !history)
      sam_error("no history available for canl generation", NULL, 2);
   if (!is_sem_elem_constr_alive(historyForCanl->constraint))
      sam_error("historyForCanl->constraint is not alive", NULL, 2);

   arity = historyForCanl->ele_nr;

   /******************************************/
   /* allocate memory for left side of canl: */
   /******************************************/
   canl_memory_left = (WORD*) MALLOC(arity * sizeof(WORD));
   if (canl_memory_left == NULL) {
      sam_error("generate_canl: Could not allocate memory for canl_memory_left.", 
		NULL, 2);
   }
   if (ADDRESS_OUT_OF_RANGE(canl_memory_left)) {
	sam_error("generate_canl: address for antilemma_memory_left out of range",
		  canl_memory_left, 3);
   }
   /**************************************************/
   /* copy variable pointers into left side of canl: */
   /**************************************************/
   GENPTR(*canl_memory_left, historyForCanl->constraint->var);
   tmpHistory = history;
   for ( i = 1 ; i < arity ; i++ ) {
      GENPTR(*(canl_memory_left + i), 
	     tmpHistory->constraint->var);
      tmpHistory = tmpHistory->previous;
   }

   /*******************************************/
   /* allocate memory for right side of canl: */
   /*******************************************/
   canl_memory_right = (WORD*) MALLOC((arity + 1) * sizeof(WORD));
   if (canl_memory_right == NULL) {
      sam_error("generate_canl: Could not allocate memory for canl_memory_right.", 
		NULL, 2);
   }
   if (ADDRESS_OUT_OF_RANGE(canl_memory_right)) {
	sam_error("generate_canl: address for antilemma_memory_right out of range",
		  canl_memory_right, 3);
   }
   /*********************************************************/
   /* Initialize pointer to list of copied terms with NULL: */
   /*********************************************************/
   next = canl_memory_right + arity;
   GENPTR(*next, NULL);
   /********************************************************/
   /* copy terms or term pointers into right side of canl: */
   /********************************************************/
   rightSide = deref(historyForCanl->constraint->right_side, bp);
   switch (GETTAG(*rightSide)) {
   case T_CONST:
	 *canl_memory_right = *rightSide;
	 break;
   case T_FVAR:
   case T_GTERM:
	 GENPTR(*canl_memory_right, rightSide);
	 break;
   case T_NGTERM:
   case T_CRTERM:
	 term = copy_term(rightSide, &next, NULL, TRUE);
	 if (term == NULL)
	    sam_error("generate_canl: Could not allocate memory for rightSide.", 
		      NULL, 2);
	 GENPTR(*canl_memory_right, term);
	 break;
   default:       
	 sam_error("generate_canl: illegal tag", rightSide, 3);
   }
   tmpHistory = history;
   for ( i = 1 ; i < arity ; i++ ) {
      rightSide = deref(tmpHistory->constraint->right_side, bp);
      switch (GETTAG(*rightSide)) {
      case T_CONST:
	    *(canl_memory_right + i) = *rightSide;
	    break;
      case T_FVAR:
      case T_GTERM:
	    GENPTR(*(canl_memory_right + i), rightSide);
	    break;
      case T_NGTERM:
      case T_CRTERM:
	    term = copy_term(rightSide, &next, NULL, TRUE);
	    if (term == NULL)
	       sam_error("generate_canl: Could not allocate memory for rightSide.", 
			 NULL, 2);
	    GENPTR(*(canl_memory_right + i), term);
	    break;
      default:       
	    sam_error("generate_canl: illegal tag", rightSide, 3);
      }
      tmpHistory = tmpHistory->previous;
   }

   /**********************************************/
   /* allocate memory for control block of canl: */
   /**********************************************/
   canl_memory = (a_list*) MALLOC(sizeof(a_list));
   if (canl_memory==NULL) {
      sam_error("generate_canl: Could not allocate memory for canl_memory.",
		NULL, 2);
   }
   if (ADDRESS_OUT_OF_RANGE(canl_memory)) {
      sam_error("generate_canl: address for canl_memory out of range",
		(WORD*) canl_memory, 3);
   } 
   /************************************************************/
   /* fill control block of canl and insert it into canl list: */
   /************************************************************/
   canl_memory->active = 1;
   canl_memory->arity = arity;
   canl_memory->left = canl_memory_left;
   canl_memory->right = canl_memory_right;
   canl_memory->infcount = 0;
   canl_memory->next = canl_list;

   return canl_memory;
}
