/******************************************************/
/*    S E T H E O                                     */
/*                                                    */
/* FILE: i_disconstr.c                                */
/* VERSION:                                           */
/* DATE:                                              */
/* AUTHOR:                                            */
/* NAME OF FILE:                                      */
/* DESCR:                                             */
/* MOD:                                               */
/* BUGS:                                              */
/******************************************************/

#include <stdio.h>
#include "tags.h"
#include "machine.h"
#include "deref.h"
#include "errors.h"
#include "disp.h"
#include "unitlemma.h"
#include "sem_constr.h"
#include "univ_constr.h"

#define INSTR_LENGTH		2
/* #define LOCAL_DEBUG */



/*************** procedures for the listhandling ******************/

int is_list(list)
WORD* list;
{
  return ( ISSYMBOL(*list) && ((GETSYMBOL(*list) == LISTFUNCT) || (GETSYMBOL(*list) == EMPTYLIST)));
}



void check_list(list)
/* Is list a list? If not, signal an error */
WORD* list;
{
  if ( !is_list(list) )
    sam_error("disconstr: Wrong argument format (list expected)", list, 3);
}



void check_nonempty_list(list)
/* Is list a non-empty list? If not, signal an error */
WORD* list;
{
  if ( !(ISSYMBOL(*list) && (GETSYMBOL(*list) == LISTFUNCT)) )
    sam_error("disconstr: Wrong argument format (non-empty list expected)", list, 3);
}



int is_list_empty(list)
WORD* list;
{
  if (! is_list(list) )
    sam_error("disconstr (is_list_empty): argument not a list", list, 3);
  return ( ISSYMBOL(*list) && (GETSYMBOL(*list) == EMPTYLIST) );
}



int list_length(list) 
WORD* list;
{
  int length = 0;
  while (!is_list_empty(list)) {
    length++;
    list = deref(list+2, bp); /* cdr */
  }
  return length;
}



void extract_constraint_sides(constraint, leftListPtr, rightListPtr)
/* constraint must be a list containing (exactly) two non-empty lists,
 * if not: error! Extract these an put 'em in leftListPtr and rightListPtr */
WORD*  constraint;
WORD** leftListPtr;
WORD** rightListPtr;
{
  check_nonempty_list(constraint); /* header correct? */
  *leftListPtr = deref(constraint+1, bp); /* car of constraint (first list element) */
  check_nonempty_list(*leftListPtr);
  constraint = deref(constraint+2, bp); /* cdr of constraint */
  check_nonempty_list(constraint); /* header correct? */
  *rightListPtr = deref(constraint+1, bp); /* car (second list element) */
  check_nonempty_list(*rightListPtr);
  
  constraint = deref(constraint+2, bp); /* cdr */
  /* constraint must be empty now 
   * (i.e. there are exactly two elements in the list) */
  if ( !(ISSYMBOL(*constraint) && (GETSYMBOL(*constraint) == EMPTYLIST)) )
    sam_error("disconstr: Wrong argument format (end-of-list expected)", constraint, 3);
}



/*************** creation of the sem_constr-tree ******************/

instr_result i_disconstr()
{
  WORD*            topList;
  WORD*            constraint;
  WORD*            leftList;
  WORD*            rightList;
  sem_junct*       rootJunct;
  sem_junct*       curJunct;
  int              mySuccess = FALSE;

  topList = deref(ARGV(0), bp);
  
  check_list(topList);
    
  /* create rootJunct on the constraint-stack
   * rootJunct is the disjunctor between the equality-subconstraints */ 
  rootJunct = add_new_sem_junct(NULL, JUNCT_DISJ); /* parent_junct is NULL */
    
  /******************************************************************/
  /* Process elements of topList:                                   */
  /******************************************************************/
    
  while (GETSYMBOL(*topList) == LISTFUNCT) {
    
    constraint = deref(topList+1, bp);    
    
    /* constraint has to be a list with two elements,
     * that are nonempty lists themselves 
     * Extract these two lists and write them into
     * leftList and rightList */
    extract_constraint_sides(constraint, &leftList, &rightList);
          
    if (list_length(leftList) != list_length(rightList))
      sam_error("disconstr: The constraint-sides are unbalanced", NULL, 2);

    pu_tr = pu_trail;

    if (un_protoco(leftList, bp, rightList, bp, 1)) {
      if (pu_tr != pu_trail) {
	/* create curJunct on the constraint-stack
	 * curJunct is the conjunctor between the elementary-constraints */ 
	curJunct = add_new_sem_junct(rootJunct, JUNCT_CONJ); /* parent_junct is the rootJunct */
	generate_sem_elem_constraints_from_pu_trail(curJunct, ELEM_EQU);
      }
      mySuccess = TRUE;
    }
    else
       pu_untrail();
    
    topList = deref(topList+2, bp); /* next element of topList */
    check_list(topList);
  }

  if (! mySuccess)
    return failure;

  if (rootJunct->arity == 0)
    /* remove rootJunct, as no constraints were generated */
    c_sp = (WORD*) rootJunct;
  else {
#ifdef LOCAL_DEBUG
    print_sem_constr_tree(rootJunct, 0);
    printf("\n");
    fflush(stdout);
#endif /* LOCAL_DEBUG */
  }

  /* try deterministic instantiation if root node has arity 1
     (root node is disjunctive) */ 
  if (rootJunct->arity == 1  &&  try_det_inst(rootJunct) == 0)
     return failure;

  pc +=INSTR_LENGTH;
  return success;
}
